"""Implementing Algebraic Reasoning (AR)."""
# pyright: ignore

from __future__ import annotations

import os
from typing import TYPE_CHECKING, TypeAlias, cast

import numpy as np
import scipy.optimize as opt  # type: ignore
from sympy import Eq  # type: ignore
from sympy import solve as sp_solve  # type: ignore
from sympy.core.backend import Symbol as SympySymbol  # type: ignore

if TYPE_CHECKING:
    from sympy import Expr  # type: ignore

    from newclid.predicates import Predicate
else:
    from symengine.lib.symengine_wrapper import Expr  # type: ignore

from newclid.numerical import ATOM

EqDict: TypeAlias = dict[SympySymbol, Expr]

# MARK: gotchas
# Everywhere we use sp.Expr, they represent an expression that's constant


os.environ.setdefault("USE_SYMENGINE", "1")


class ARTable:
    """The coefficient matrix."""

    def __init__(self):
        self.v2e: EqDict = {}  # the table {var: {vark : coefk}} var = sum coefk*vark

        # for why (linprog)
        self._c = np.zeros((0))
        self._v2i: dict[SympySymbol, int] = {}  # v -> index of row in A.
        self.predicates: list[Predicate] = []  # equal number of columns.
        self._mA: np.ndarray[tuple[int, ...], np.dtype[np.float64]] = np.zeros((0, 0))

    def expr_delta(self, vc: Expr) -> bool:
        """
        There is only constant delta between vc and the system
        """
        sub_result = self.substitute_in_existing_expressions(vc)
        is_constant = len(sub_result.free_symbols) == 0
        return is_constant

    def substitute_in_existing_expressions(self, vc: Expr) -> Expr:
        before_simp = vc.xreplace(self.v2e)
        after_simp = cast(Expr, before_simp.expand())
        return after_simp

    def add_expr(self, expression: Expr, predicate: Predicate) -> bool:
        """
        Add a new equality (sum cv = 0), represented by the list of tuples vc=[(v, c), ..].
        Return True iff the equality can already be deduced by the internal system
        """
        simplified_expression = self.substitute_in_existing_expressions(expression)

        if len(simplified_expression.free_symbols) == 0:
            return False

        new_vars: set[SympySymbol] = simplified_expression.free_symbols - set(
            self.v2e.keys()
        )

        if len(new_vars) == 0:
            v, e = _reconcile(simplified_expression)
            if self.v2e[v] != v:
                assert False, (
                    f"{expression=} simplified to {simplified_expression=}, {v=}, {e=}. Previously, {self.v2e[v]=}. "
                    f"This should never happen because we should have already reconciled this variable above."
                )
            self._replace(v, e)

        else:
            # We have at least one new variable.
            assert simplified_expression != 0

            for v in new_vars:
                self._add_free(v)

            # Select one new variable to be dependent.
            dependent_variable = new_vars.pop()
            expr = cast(
                list[Expr], sp_solve(Eq(simplified_expression, 0), dependent_variable)
            )[0]

            self.v2e[dependent_variable] = expr

        self._register(simplified_expression, predicate)
        return True

    def why_expr(self, vc: Expr) -> tuple[Predicate, ...]:
        """AR traceback == MILP."""
        # why expr == 0?
        # Solve min(c^Tx) s.t. A_eq * x = b_eq, x >= 0
        vc = vc.expand()
        if vc == 0:
            no_justification: list[Predicate] = []
            return tuple(no_justification)

        b_eq = np.array([0] * len(self._v2i))
        for v in cast(set[SympySymbol], vc.free_symbols):
            b_eq[self._v2i[v]] += float(vc.coeff(v))

        try:
            x = opt.linprog(c=self._c, A_eq=self._mA, b_eq=b_eq, method="highs")["x"]
        except ValueError:
            x = opt.linprog(c=self._c, A_eq=self._mA, b_eq=b_eq)["x"]

        if x is None:
            ar_lines = "\n".join(f"{k} = {v}" for k, v in self.v2e.items())
            raise ValueError(
                f"Could not fild a traceback through AR for expression {vc} = 0.\nAR table:\n{ar_lines}"
            )

        predicates: list[Predicate] = []
        for i, dep in enumerate(self.predicates):
            if x[2 * i] > ATOM or x[2 * i + 1] > ATOM:
                if dep not in predicates:
                    predicates.append(dep)
        return tuple(predicates)

    def _register(self, vc: Expr, predicate: Predicate) -> None:
        """Register a new equality vc=[(v, c), ..] with traceback dependency dep."""
        vc = cast(Expr, vc.expand())
        if vc == 0:
            return

        for v in cast(set[SympySymbol], vc.free_symbols):
            if v not in self._v2i:
                self._v2i[v] = len(self._v2i)

        (m, n), length = self._mA.shape, len(self._v2i)
        if length > m:
            self._mA = np.concatenate([self._mA, np.zeros([length - m, n])], 0)

        new_column = np.zeros((len(self._v2i), 2))  # N, 2
        for v, c in _vars_and_coeffs(vc).items():
            new_column[self._v2i[v], 0] += c
            new_column[self._v2i[v], 1] -= c

        self._mA = np.concatenate((self._mA, new_column), 1)
        self._c = np.concatenate((self._c, np.array([1.0, -1.0])))
        self.predicates += [predicate]

    def _add_free(self, v: SympySymbol) -> None:
        self.v2e[v] = v

    def _replace(self, v0: SympySymbol, e0: Expr) -> None:
        for v, e in list(self.v2e.items()):
            self.v2e[v] = e.subs(v0, e0).expand()


def _reconcile(e: Expr) -> tuple[SympySymbol, Expr]:
    """Reconcile one variable in the expression e=0, given const."""
    e = e.expand()
    free_vars = set(cast(set[SympySymbol], e.free_symbols))
    assert len(free_vars) > 0, f"{e=} should have at least two free variables"
    v0 = free_vars.pop()
    return v0, sp_solve(e, v0)[0]


def _vars_and_coeffs(expr: Expr) -> dict[SympySymbol, float]:
    ret: dict[SympySymbol, float] = {}
    for v in expr.free_symbols:
        assert isinstance(v, SympySymbol)
        coeff: SympySymbol = expr.coeff(v)
        assert coeff.is_number
        ret[v] = float(coeff)
    return ret



"""Module to convert problems to and from LLM input."""

from __future__ import annotations

import re
from typing import Self

import numpy as np
from pydantic import BaseModel

from newclid.api import GeometricSolverBuilder
from newclid.jgex.clause import JGEXClause, JGEXConstruction
from newclid.jgex.formulation import JGEXFormulation
from newclid.jgex.jgex_setup_data import (
    JGEXClauseInProof,
    JGEXSetupData,
    PredicateConstructionInSetup,
    jgex_clauses_to_setup_data,
)
from newclid.jgex.problem_builder import JGEXProblemBuilder
from newclid.justifications._index import JustificationType
from newclid.predicate_types import PredicateArgument
from newclid.problem import PredicateConstruction, predicate_to_construction
from newclid.proof_data import PredicateInProof, ProofData, ProofStep


def new_problem_from_llm_aux_output(
    initial_problem: JGEXFormulation, aux_output: str, aux_tag: str
) -> JGEXFormulation:
    sanitized_aux_output = _clauses_in_aux_output(aux_output, aux_tag)
    aux_clauses = tuple(_llm_str_to_clause(c) for c in sanitized_aux_output)
    problem_with_aux = JGEXFormulation(
        name=initial_problem.name + "_with_" + aux_output,
        setup_clauses=initial_problem.setup_clauses,
        auxiliary_clauses=aux_clauses,
        goals=initial_problem.goals,
    )
    return problem_with_aux


def problem_to_llm_input(
    problem: JGEXFormulation,
    aux_tag: str,
    rng: np.random.Generator | None = None,
    max_attempts: int = 200,
) -> str:
    """
    Convert a problem to a string that can be used as input for the LLM.

    Args:
        problem: The problem to convert.
        aux_tag: The tag to use for the aux clauses.
        rng: The random number generator to use.
        max_attempts: The maximum number of attempts to make for building the problem.
    """
    formulation_with_aux_in_setup = JGEXFormulation(
        name=problem.name,
        setup_clauses=problem.setup_clauses + problem.auxiliary_clauses,
        auxiliary_clauses=(),
        goals=problem.goals,
    )
    setup_data = _problem_jgex_to_setup_data(
        formulation_with_aux_in_setup, rng=rng, max_attempts=max_attempts
    )
    training_data = AuxTrainingDatapoint.from_setup_data(setup_data, aux_tag)
    return training_data.input


def problem_to_llm_input_without_predicates(problem: JGEXFormulation) -> str:
    formulation_with_aux_in_setup = JGEXFormulation(
        name=problem.name,
        setup_clauses=problem.setup_clauses + problem.auxiliary_clauses,
        auxiliary_clauses=(),
        goals=problem.goals,
    )
    return str(formulation_with_aux_in_setup)


class TrainingDatapoint(BaseModel):
    aux_io: list[AuxTrainingDatapoint]
    """Aux input and output."""
    proof_output: str
    """Proof to predict."""

    @classmethod
    def from_proof_data(
        cls, setup_data: JGEXSetupData, proof_data: ProofData, aux_tag: str
    ) -> Self:
        return cls(
            aux_io=[AuxTrainingDatapoint.from_setup_data(setup_data, aux_tag)],
            proof_output=_join_proof_steps(proof_data.proof_steps),
        )

    @classmethod
    def from_proof_data_aux_combinations(
        cls, setup_data: JGEXSetupData, proof_data: ProofData, aux_tag: str
    ) -> Self:
        return cls(
            aux_io=AuxTrainingDatapoint.from_setup_data_aux_combinations(
                setup_data, aux_tag
            ),
            proof_output=_join_proof_steps(proof_data.proof_steps),
        )


class AuxTrainingDatapoint(BaseModel):
    input: str
    """Setup and given aux clauses plus the goal."""
    aux_output: str
    """Aux constructions plus theirs predicates to predict."""

    @classmethod
    def from_setup_data(cls, setup_data: JGEXSetupData, aux_tag: str) -> Self:
        goals_str = _join_goals(setup_data.goals)
        return cls(
            input=_join_clauses(setup_data.setup_clauses) + goals_str,
            aux_output=_join_clauses(setup_data.aux_clauses, tag=aux_tag),
        )

    @classmethod
    def from_setup_data_aux_combinations(
        cls, setup_data: JGEXSetupData, aux_tag: str
    ) -> list[Self]:
        goals_str = _join_goals(setup_data.goals)
        aux_clauses = setup_data.aux_clauses

        aux_io: list[Self] = []
        for aux_left_out_index in range(1, len(aux_clauses) + 1):
            aux_included = aux_clauses[:-aux_left_out_index]
            aux_to_predict = aux_clauses[len(aux_clauses) - aux_left_out_index :]
            aux_io.append(
                cls(
                    input=_join_clauses(setup_data.setup_clauses + aux_included)
                    + goals_str,
                    aux_output=_join_clauses(aux_to_predict, tag=aux_tag),
                )
            )
        return aux_io


def _problem_jgex_to_setup_data(
    problem: JGEXFormulation,
    max_attempts: int,
    aux_clauses: list[JGEXClause] | None = None,
    rng: np.random.Generator | None = None,
) -> JGEXSetupData:
    if aux_clauses is None:
        aux_clauses = []
    setup_clauses = [
        clause for clause in problem.setup_clauses if clause not in aux_clauses
    ]
    problem_builder = JGEXProblemBuilder(rng=rng).with_problem(problem)
    jgex_problem_setup = problem_builder.build(
        max_attempts_to_satisfy_goals_numerically=max_attempts
    )
    GeometricSolverBuilder(rng=rng).build(jgex_problem_setup)
    setup_data, _ = jgex_clauses_to_setup_data(
        setup_clauses=setup_clauses,
        aux_clauses=aux_clauses,
        clauses_consequences=problem_builder.clauses_consequences,
        goals=list(problem.goals),
    )
    return setup_data


def _join_goals(goals: list[PredicateConstructionInSetup]) -> str:
    return " ? " + ", ".join(
        _predicate_construction_in_setup_in_proof_to_llm_str(g) for g in goals
    )


def _join_clauses(clauses: list[JGEXClauseInProof], tag: str = "") -> str:
    return "; ".join(_clause_in_proof_to_llm_str(c, tag) for c in clauses)


def _join_proof_steps(proof_steps: list[ProofStep]) -> str:
    return ", ".join(_proof_step_to_llm_str(step) for step in proof_steps)


def _predicate_construction_in_setup_in_proof_to_llm_str(
    predicate_in_prood: PredicateConstructionInSetup,
) -> str:
    return f"[{predicate_in_prood.id}] {_predicate_construction_to_llm_str(predicate_in_prood.construction)}"


def _predicate_in_proof_to_llm_str(
    predicate_in_prood: PredicateInProof,
) -> str:
    return f"[{predicate_in_prood.id}] {_predicate_construction_to_llm_str(predicate_to_construction(predicate_in_prood.predicate))}"


def _predicate_construction_to_llm_str(construction: PredicateConstruction) -> str:
    return (
        construction.predicate_type.value
        + " "
        + " ".join(str(arg) for arg in construction.args)
    )


def _clause_in_proof_to_llm_str(clause: JGEXClauseInProof, tag: str = "") -> str:
    points = " ".join(clause.added_points)
    constructions = ", ".join(
        _jgex_construction_to_llm_str(c) for c in clause.constructions
    )
    if tag:
        tag = tag.strip() + " "
    predicates = ", ".join(
        _predicate_construction_in_setup_in_proof_to_llm_str(pred_construction_in_setup)
        for pred_construction_in_setup in clause.predicates
    )
    return f"{tag}{points} = {constructions} ({predicates})"


def _llm_str_to_clause(llm_str: str) -> JGEXClause:
    points_str, constructions_str = llm_str.split(" = ")
    points = tuple(PredicateArgument(p) for p in points_str.split(" "))
    constructions = tuple(
        _llm_str_to_construction(c_str) for c_str in constructions_str.split(", ")
    )
    return JGEXClause(points=points, constructions=constructions)


def _clauses_in_aux_output(aux_output: str, aux_tag: str) -> list[str]:
    without_aux_tag = aux_output.replace(aux_tag, "")
    without_predicates = re.sub(r"\s*\([^)]*\)", "", without_aux_tag).strip()
    return [s.strip() for s in without_predicates.split(";")]


def _proof_step_to_llm_str(step: ProofStep) -> str:
    premises_str = f"[{','.join(step.applied_on_predicates)}]"
    if step.justification.dependency_type == JustificationType.RULE_APPLICATION:
        reason_str = step.justification.rule.id
    else:
        reason_str = step.justification.dependency_type.value
    return f"{_predicate_in_proof_to_llm_str(step.proven_predicate)} <({reason_str})= {premises_str}"


def _llm_str_to_construction(llm_str: str) -> JGEXConstruction:
    return JGEXConstruction.from_str(llm_str)


def _jgex_construction_to_llm_str(construction: JGEXConstruction) -> str:
    return construction.name + " " + " ".join(str(arg) for arg in construction.args)



from __future__ import annotations

import numpy as np
from matplotlib.axes import Axes

from newclid.draw.theme import DrawTheme


class TextAnimation:
    def __init__(
        self,
        ax: Axes,
        proof_lines: list[str],
        theme: DrawTheme,
        n_lines_shown: int = 13,
        fontsize: int = 16,
    ):
        if n_lines_shown % 2 == 0:
            raise ValueError(
                "n_lines_shown must be odd for the current line to be highlighted in the center"
            )
        self.ax = ax
        self.proof_lines = proof_lines
        self.theme = theme
        self.n_lines_shown = n_lines_shown
        self.highlighted_line_index = self.n_lines_shown // 2
        self.current_line_index = 0
        self.shown_after_current_line = self.highlighted_line_index
        self.shown_before_current_line = n_lines_shown - self.shown_after_current_line
        self.fontsize = fontsize

    @property
    def current_lines(self) -> list[str]:
        min_index = max(self.current_line_index - self.shown_before_current_line + 1, 0)
        current_and_past = self.proof_lines[min_index : self.current_line_index + 1]
        max_index = min(
            self.current_line_index + self.shown_after_current_line + 1,
            len(self.proof_lines),
        )
        lines_after = self.proof_lines[self.current_line_index + 1 : max_index]
        missing_before_lines = max(
            0, self.highlighted_line_index + 1 - len(current_and_past)
        )
        missing_after_lines = max(0, self.shown_after_current_line - len(lines_after))
        padded_lines = (
            [""] * missing_before_lines
            + current_and_past
            + lines_after
            + [""] * missing_after_lines
        )
        assert len(padded_lines) == self.n_lines_shown, (
            f"Expected {self.n_lines_shown} lines, got {len(padded_lines)}."
            f"lines_before: {len(current_and_past)}, lines_after: {len(lines_after)},"
            f"line_padded: {missing_before_lines}, current_line_index: {self.current_line_index},"
            f"shown_before_current_line: {self.shown_before_current_line},"
            f"shown_after_current_line: {self.shown_after_current_line},"
            f"highlighted_line_index: {self.highlighted_line_index}."
        )
        return padded_lines

    def init(self):
        self.ax.set_facecolor((0.0, 0.0, 0.0))
        self.ax.get_xaxis().set_visible(False)
        self.ax.get_yaxis().set_visible(False)

        self.current_line_index = 0
        text_ys = np.linspace(0, 1, self.n_lines_shown)

        self.lines_artists = [
            self.ax.text(  # pyright: ignore
                1,
                text_ys[i],
                "",
                ha="right",
                va="baseline",
                color=self.theme.text_color,
                family=["DejaVu Sans", "STIXGeneral"],
                fontsize=self.fontsize,
                transform=self.ax.transAxes,
            )
            for i in range(self.n_lines_shown)[::-1]
        ]
        current_line_artist = self.lines_artists[self.highlighted_line_index]
        current_line_artist.set_color(self.theme.point_color)

        for i, artist in enumerate(self.lines_artists):
            size = int(
                _parabola(i, self.highlighted_line_index, max=1, min=0.3)
                * self.fontsize
            )
            artist.set_fontsize(size)
            alpha = _parabola(i, self.highlighted_line_index, max=1, min=0.5)
            artist.set_alpha(alpha)
        return self.lines_artists

    def update(self, frame: int):
        for artist, line in zip(self.lines_artists, self.current_lines):
            artist.set_text(line)
        self.current_line_index += 1
        return self.lines_artists


def _parabola(x: float, h: float, max: float, min: float) -> float:
    return max - ((max - min) / h**2) * (x - h) ** 2



from pathlib import Path

import numpy as np
import pytest
from newclid.jgex.formulation import JGEXFormulation, jgex_formulation_from_txt_file
from newclid.llm_input import problem_to_llm_input

ALL_JGEX_COMPATIBLE_PROBLEM_FILES = [
    "examples.txt",
    "large_examples.txt",
    "imo.txt",
    "imo_sl.txt",
]

PROBLEMS_EXPECTED_TO_FAIL_TO_BUILD = [
    "breaking_cc_tangent",
    "breaking_cc_tangent_by_kiss",
    "breaking_cc_itangent",
    "breaking_cc_itangent_by_kiss",
    "2020_p1",
    # ag-shortlist
    "2018_sl_g4a",
    # non-ag 30
    "2005_p1",
    "2007_p2",
    "2013_p3",
]

ALL_PROBLEMS: list[JGEXFormulation] = []
for file_name in ALL_JGEX_COMPATIBLE_PROBLEM_FILES:
    ALL_PROBLEMS.extend(
        jgex_formulation_from_txt_file(
            Path(__file__).parent.parent.joinpath("problems_datasets", file_name)
        ).values()
    )


@pytest.mark.parametrize(
    "jgex_problem", ALL_PROBLEMS, ids=[p.name for p in ALL_PROBLEMS]
)
def test_can_build_all_example_problems(jgex_problem: JGEXFormulation):
    if jgex_problem.name in PROBLEMS_EXPECTED_TO_FAIL_TO_BUILD:
        pytest.skip("This problem is expected to fail to build.")
    rng = np.random.default_rng(42)
    problem_to_llm_input(jgex_problem, rng=rng, aux_tag="!aux")



"""Definitions of points at the intersection/tangent of lines and circles."""

from newclid.jgex.clause import JGEXClause
from newclid.jgex.constructions._index import JGEXConstructionName
from newclid.jgex.definition import JGEXDefinition, SketchConstruction, v
from newclid.jgex.sketch import (
    sketch_cc_itangent,
    sketch_cc_tangent,
    sketch_circle,
    sketch_function_name,
    sketch_line,
    sketch_pline,
    sketch_tangent,
    sketch_tline,
)

# Line Intersections

INTERSECTION_LINE_LINE = JGEXDefinition(
    name=JGEXConstructionName.INTERSECTION_LINE_LINE,
    args=(v("x"), v("a"), v("b"), v("c"), v("d")),
    rely_on_points={v("x"): (v("a"), v("b"), v("c"), v("d"))},
    requirements=JGEXClause.from_str("npara a b c d, ncoll a b c d")[0],
    clauses=JGEXClause.from_str("x : coll x a b, coll x c d"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_line.__name__), args=(v("a"), v("b"))
        ),
        SketchConstruction(
            name=sketch_function_name(sketch_line.__name__), args=(v("c"), v("d"))
        ),
    ),
    input_points=(v("a"), v("b"), v("c"), v("d")),
    output_points=(v("x"),),
)

INTERSECTION_LINE_PARA = JGEXDefinition(
    name=JGEXConstructionName.INTERSECTION_LINE_PARA,
    args=(v("x"), v("a"), v("b"), v("c"), v("m"), v("n")),
    rely_on_points={v("x"): (v("a"), v("b"), v("c"), v("m"), v("n"))},
    requirements=JGEXClause.from_str("npara m n a b, ncoll a b c, ncoll c m n")[0],
    clauses=JGEXClause.from_str("x : coll x a b, para c x m n"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_line.__name__), args=(v("a"), v("b"))
        ),
        SketchConstruction(
            name=sketch_function_name(sketch_pline.__name__),
            args=(v("c"), v("m"), v("n")),
        ),
    ),
    input_points=(v("a"), v("b"), v("c"), v("m"), v("n")),
    output_points=(v("x"),),
)

INTERSECTION_PARA_PARA = JGEXDefinition(
    name=JGEXConstructionName.INTERSECTION_PARA_PARA,
    args=(v("x"), v("a"), v("b"), v("c"), v("d"), v("e"), v("f")),
    rely_on_points={v("x"): (v("a"), v("b"), v("c"), v("d"), v("e"), v("f"))},
    requirements=JGEXClause.from_str("diff a d, npara b c e f")[0],
    clauses=JGEXClause.from_str("x : para x a b c, para x d e f"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_pline.__name__),
            args=(v("a"), v("b"), v("c")),
        ),
        SketchConstruction(
            name=sketch_function_name(sketch_pline.__name__),
            args=(v("d"), v("e"), v("f")),
        ),
    ),
    input_points=(v("a"), v("b"), v("c"), v("d"), v("e"), v("f")),
    output_points=(v("x"),),
)


INTERSECTION_LINE_PERP = JGEXDefinition(
    name=JGEXConstructionName.INTERSECTION_LINE_PERP,
    args=(v("x"), v("a"), v("b"), v("c"), v("d"), v("e")),
    rely_on_points={v("x"): (v("a"), v("b"), v("c"), v("d"), v("e"))},
    requirements=JGEXClause.from_str("ncoll a b c, nperp a b d e")[0],
    clauses=JGEXClause.from_str("x : coll x a b, perp x c d e"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_line.__name__), args=(v("a"), v("b"))
        ),
        SketchConstruction(
            name=sketch_function_name(sketch_tline.__name__),
            args=(v("c"), v("d"), v("e")),
        ),
    ),
    input_points=(v("a"), v("b"), v("c"), v("d"), v("e")),
    output_points=(v("x"),),
)

INTERSECTION_PERP_PERP = JGEXDefinition(
    name=JGEXConstructionName.INTERSECTION_PERP_PERP,
    args=(v("x"), v("a"), v("b"), v("c"), v("d"), v("e"), v("f")),
    rely_on_points={v("x"): (v("a"), v("b"), v("c"), v("d"), v("e"), v("f"))},
    requirements=JGEXClause.from_str("diff a d, npara b c e f")[0],
    clauses=JGEXClause.from_str("x : perp x a b c, perp x d e f"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_tline.__name__),
            args=(v("a"), v("b"), v("c")),
        ),
        SketchConstruction(
            name=sketch_function_name(sketch_tline.__name__),
            args=(v("d"), v("e"), v("f")),
        ),
    ),
    input_points=(v("a"), v("b"), v("c"), v("d"), v("e"), v("f")),
    output_points=(v("x"),),
)

# Circle Intersections

INTERSECTION_LINE_CIRCLE = JGEXDefinition(
    name=JGEXConstructionName.INTERSECTION_LINE_CIRCLE,
    args=(v("x"), v("a"), v("o"), v("b")),
    rely_on_points={v("x"): (v("a"), v("o"), v("b"))},
    requirements=JGEXClause.from_str("diff a b, diff o b, nperp b o b a")[0],
    clauses=JGEXClause.from_str("x : coll x a b, cong o b o x"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_line.__name__), args=(v("b"), v("a"))
        ),
        SketchConstruction(
            name=sketch_function_name(sketch_circle.__name__),
            args=(v("o"), v("o"), v("b")),
        ),
    ),
    input_points=(v("a"), v("o"), v("b")),
    output_points=(v("x"),),
)

INTERSECTION_CIRCLE_CIRCLE = JGEXDefinition(
    name=JGEXConstructionName.INTERSECTION_CIRCLE_CIRCLE,
    args=(v("x"), v("o"), v("w"), v("a")),
    rely_on_points={v("x"): (v("o"), v("w"), v("a"))},
    requirements=JGEXClause.from_str("ncoll o w a")[0],
    clauses=JGEXClause.from_str("x : cong o a o x, cong w a w x"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_circle.__name__),
            args=(v("o"), v("o"), v("a")),
        ),
        SketchConstruction(
            name=sketch_function_name(sketch_circle.__name__),
            args=(v("w"), v("w"), v("a")),
        ),
    ),
    input_points=(v("o"), v("w"), v("a")),
    output_points=(v("x"),),
)

# Tangents

TANGENT = JGEXDefinition(
    name=JGEXConstructionName.TANGENT,
    args=(v("x"), v("y"), v("a"), v("o"), v("b")),
    rely_on_points={
        v("x"): (v("o"), v("a"), v("b")),
        v("y"): (v("o"), v("a"), v("b")),
    },
    requirements=JGEXClause.from_str("diff o a, diff o b, diff a b")[0],
    clauses=JGEXClause.from_str(
        "x : cong o x o b, perp a x o x; y : cong o y o b, perp a y o y"
    ),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_tangent.__name__),
            args=(v("a"), v("o"), v("b")),
        ),
    ),
    input_points=(v("a"), v("o"), v("b")),
    output_points=(v("x"), v("y")),
)

LINE_CIRCLE_TANGENT = JGEXDefinition(
    name=JGEXConstructionName.LINE_CIRCLE_TANGENT,
    args=(v("x"), v("a"), v("o")),
    rely_on_points={v("x"): (v("a"), v("o"))},
    requirements=JGEXClause.from_str("diff a o")[0],
    clauses=JGEXClause.from_str("x : perp a x a o"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_tline.__name__),
            args=(v("a"), v("a"), v("o")),
        ),
    ),
    input_points=(v("a"), v("o")),
    output_points=(v("x"),),
)


CIRCLE_CIRCLE_TANGENT = JGEXDefinition(
    name=JGEXConstructionName.CIRCLE_CIRCLE_TANGENT,
    args=(v("x"), v("y"), v("z"), v("i"), v("o"), v("a"), v("w"), v("b")),
    rely_on_points={
        v("x"): (v("o"), v("a"), v("w"), v("b"), v("y")),
        v("y"): (v("o"), v("a"), v("w"), v("b"), v("x")),
        v("z"): (v("o"), v("a"), v("w"), v("b"), v("i")),
        v("i"): (v("o"), v("a"), v("w"), v("b"), v("z")),
    },
    requirements=JGEXClause.from_str("diff o a, diff w b, diff o w")[0],
    clauses=JGEXClause.from_str(
        "x y : cong o x o a, cong w y w b, perp x o x y, perp y w y x; z i : cong o z o a, cong w i w b, perp z o z i, perp i w i z"
    ),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_cc_tangent.__name__),
            args=(v("o"), v("a"), v("w"), v("b")),
        ),
    ),
    input_points=(v("o"), v("a"), v("w"), v("b")),
    output_points=(v("x"), v("y"), v("z"), v("i")),
)


CIRCLE_CIRCLE_ITANGENT = JGEXDefinition(
    name=JGEXConstructionName.CIRCLE_CIRCLE_ITANGENT,
    args=(v("x"), v("y"), v("z"), v("i"), v("o"), v("a"), v("w"), v("b")),
    rely_on_points={
        v("x"): (v("o"), v("a"), v("w"), v("b"), v("y")),
        v("y"): (v("o"), v("a"), v("w"), v("b"), v("x")),
        v("z"): (v("o"), v("a"), v("w"), v("b"), v("i")),
        v("i"): (v("o"), v("a"), v("w"), v("b"), v("z")),
    },
    requirements=JGEXClause.from_str("diff o a, diff w b, diff o w")[0],
    clauses=JGEXClause.from_str(
        "x y : cong o x o a, cong w y w b, perp x o x y, perp y w y x; z i : cong o z o a, cong w i w b, perp z o z i, perp i w i z"
    ),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_cc_itangent.__name__),
            args=(v("o"), v("a"), v("w"), v("b")),
        ),
    ),
    input_points=(v("o"), v("a"), v("w"), v("b")),
    output_points=(v("x"), v("y"), v("z"), v("i")),
)


INTERSECTION_CONSTRUCTIONS = [
    INTERSECTION_LINE_LINE,
    INTERSECTION_LINE_PARA,
    INTERSECTION_PARA_PARA,
    INTERSECTION_LINE_PERP,
    INTERSECTION_PERP_PERP,
    INTERSECTION_LINE_CIRCLE,
    INTERSECTION_CIRCLE_CIRCLE,
    TANGENT,
    LINE_CIRCLE_TANGENT,
    CIRCLE_CIRCLE_TANGENT,
    CIRCLE_CIRCLE_ITANGENT,
]



"""Graph linking predicates by justifications as edges."""

from __future__ import annotations

import logging
from pathlib import Path
from typing import TYPE_CHECKING, Collection

from pyvis.network import Network  # type: ignore

from newclid.justifications._index import JustificationType
from newclid.justifications.justification import justify_dependency
from newclid.proof_justifications import goals_justifications
from newclid.tools import add_edge, boring_predicate

if TYPE_CHECKING:
    from newclid.justifications.justification import Justification
    from newclid.predicates import Predicate
    from newclid.proof_state import ProofState


LOGGER = logging.getLogger(__name__)


class PredicatesGraph:
    """Hyper graph linking predicates by justifications as hyper-edges."""

    def __init__(self) -> None:
        self.hyper_graph: dict[Predicate, Justification] = {}

    def has_edge(self, justification: Justification):
        return self.hyper_graph.get(justification.predicate) == justification

    @property
    def predicates(self) -> list[Predicate]:
        return list(self.hyper_graph.keys())

    def premises(self) -> list[Justification]:
        premises: list[Justification] = []
        for _, justification in self.hyper_graph.items():
            if justification.dependency_type == JustificationType.ASSUMPTION:
                premises.append(justification)
        return premises

    def save_pyvis(
        self, *, proof_state: ProofState, path: Path, stars: Collection[Predicate] = []
    ):
        if stars:
            justifications, _ = goals_justifications(list(stars), proof_state)
        else:
            justifications = tuple(dep for _, dep in self.hyper_graph.items())
        net = Network("1080px", directed=True)
        for justification in justifications:
            if boring_predicate(justification.predicate):
                continue
            shape = "dot"
            color = "#97c2fc"
            if justification.predicate in stars:
                shape = "star"
                color = "gold"
            net.add_node(  # type: ignore
                str(justification.predicate),
                title=f"{justification.dependency_type.value}",
                shape=shape,
                color=color,
                size=10,
            )
        for justification in justifications:
            if boring_predicate(justification.predicate):
                continue
            for premise in justify_dependency(justification, proof_state):
                add_edge(net, str(premise), str(justification.predicate))  # type: ignore
        net.options.layout = {  # type: ignore
            "hierarchical": {
                "enabled": True,
                "direction": "LR",
                "sortMethod": "directed",
            },
        }
        net.show_buttons(filter_=["physics", "layout"])  # type: ignore
        net.show(str(path), notebook=False)  # type: ignore



from __future__ import annotations

from typing import Literal

from newclid.numerical import nearly_zero
from newclid.predicate_types import PredicateArgument
from newclid.predicates._index import PredicateType
from newclid.predicates._interface import NumericalPredicate, PredicateInterface
from newclid.symbols.points_registry import Line
from newclid.symbols.symbols_registry import LineOrCircle, SymbolsRegistry


class Perp(PredicateInterface):
    """perp A B C D -
    Represent that the line AB is perpendicular to the line CD.
    """

    predicate_type: Literal[PredicateType.PERPENDICULAR] = PredicateType.PERPENDICULAR
    line1: Line
    line2: Line

    @staticmethod
    def preparse(
        args: tuple[PredicateArgument, ...],
    ) -> tuple[PredicateArgument, ...] | None:
        a, b, c, d = args
        if a == b or c == d:
            return None
        a, b = sorted((a, b))
        c, d = sorted((c, d))
        if (a, b) > (c, d):
            a, b, c, d = c, d, a, b
        return (a, b, c, d)

    def check_numerical(self) -> bool:
        a, b = self.line1
        c, d = self.line2
        return nearly_zero((a.num - b.num).dot(c.num - d.num))

    def symbols(self, symbols: SymbolsRegistry) -> tuple[LineOrCircle, ...]:
        l1 = symbols.lines.line_thru_pair(*self.line1)
        l2 = symbols.lines.line_thru_pair(*self.line2)
        return (l1, l2)

    def to_constructive(self, point: str) -> str:
        a, b = [p.name for p in self.line1]
        c, d = [p.name for p in self.line2]
        if point in [c, d]:
            a, b, c, d = c, d, a, b
        if point == b:
            a, b = b, a
        if point == d:
            c, d = d, c
        if a == c and a == point:
            return f"on_dia {a} {b} {d}"
        return f"on_tline {a} {b} {c} {d}"

    def to_tokens(self) -> tuple[PredicateArgument, ...]:
        return tuple(PredicateArgument(p.name) for p in self.line1 + self.line2)

    def __str__(self) -> str:
        a, b = self.line1
        c, d = self.line2
        return f"{a}{b} ⟂ {c}{d}"


class NPerp(NumericalPredicate):
    """nperp A B C D -
    Represent that lines AB and CD are NOT perpendicular.

    Numerical only.
    """

    predicate_type: Literal[PredicateType.N_PERPENDICULAR] = (
        PredicateType.N_PERPENDICULAR
    )

    line1: Line
    line2: Line

    @staticmethod
    def preparse(
        args: tuple[PredicateArgument, ...],
    ) -> tuple[PredicateArgument, ...] | None:
        return Perp.preparse(args)

    def check_numerical(self) -> bool:
        return not Perp(line1=self.line1, line2=self.line2).check_numerical()

    def to_tokens(self) -> tuple[PredicateArgument, ...]:
        return tuple(PredicateArgument(p.name) for p in self.line1 + self.line2)

    def __str__(self) -> str:
        a, b = self.line1
        c, d = self.line2
        return f"{a}{b} ⟂̸ {c}{d}"



"""Define predicates used in Newclid."""

from __future__ import annotations

from typing import TYPE_CHECKING, Annotated, TypeAlias

from pydantic import Field

from newclid.predicates._index import PredicateType
from newclid.predicates._interface import PredicateInterface
from newclid.predicates.circumcenter import Circumcenter
from newclid.predicates.collinearity import Coll, NColl
from newclid.predicates.congruence import Cong
from newclid.predicates.constant_angle import ACompute, ConstantAngle
from newclid.predicates.constant_length import ConstantLength, LCompute
from newclid.predicates.constant_ratio import ConstantRatio, RCompute
from newclid.predicates.cyclic import Cyclic
from newclid.predicates.different import Diff
from newclid.predicates.equal_angles import EqAngle
from newclid.predicates.equal_ratios import EqRatio
from newclid.predicates.equation_angle import AngleEquation
from newclid.predicates.equation_length import LengthEquation
from newclid.predicates.midpoint import MidPoint
from newclid.predicates.obtuse_angle import ObtuseAngle
from newclid.predicates.parallelism import NPara, Para
from newclid.predicates.perpendicularity import NPerp, Perp
from newclid.predicates.pythagoras import PythagoreanConclusions, PythagoreanPremises
from newclid.predicates.sameclock import SameClock
from newclid.predicates.sameside import NSameSide, SameSide
from newclid.predicates.squared_constant_length import SquaredConstantLength
from newclid.predicates.squared_constant_ratio import SquaredConstantRatio
from newclid.predicates.triangles_congruent import ContriClock, ContriReflect
from newclid.predicates.triangles_similar import SimtriClock, SimtriReflect
from newclid.tools import str_to_fraction

if TYPE_CHECKING:
    from newclid.problem import PredicateConstruction
    from newclid.symbols.points_registry import PointsRegisty

SYMBOLIC_PREDICATES = {
    PredicateType.ANGLE_EQUATION: AngleEquation,
    PredicateType.COLLINEAR: Coll,
    PredicateType.CONGRUENT: Cong,
    PredicateType.LENGTH_EQUATION: LengthEquation,
    PredicateType.MIDPOINT: MidPoint,
    PredicateType.PARALLEL: Para,
    PredicateType.PERPENDICULAR: Perp,
    PredicateType.CYCLIC: Cyclic,
    PredicateType.CIRCUMCENTER: Circumcenter,
    PredicateType.EQUAL_ANGLES: EqAngle,
    PredicateType.EQUAL_RATIOS: EqRatio,
    PredicateType.CONSTANT_LENGTH: ConstantLength,
    PredicateType.CONSTANT_RATIO: ConstantRatio,
    PredicateType.CONSTANT_ANGLE: ConstantAngle,
    PredicateType.SQUARED_CONSTANT_LENGTH: SquaredConstantLength,
    PredicateType.SQUARED_CONSTANT_RATIO: SquaredConstantRatio,
    PredicateType.SIMTRI_CLOCK: SimtriClock,
    PredicateType.SIMTRI_REFLECT: SimtriReflect,
    PredicateType.CONTRI_CLOCK: ContriClock,
    PredicateType.CONTRI_REFLECT: ContriReflect,
}

NUMERICAL_PREDICATES = {
    PredicateType.DIFFERENT: Diff,
    PredicateType.N_COLLINEAR: NColl,
    PredicateType.N_PARALLEL: NPara,
    PredicateType.N_PERPENDICULAR: NPerp,
    PredicateType.N_SAME_SIDE: NSameSide,
    PredicateType.OBTUSE_ANGLE: ObtuseAngle,
    PredicateType.SAME_CLOCK: SameClock,
    PredicateType.SAME_SIDE: SameSide,
}

INTEGRATED_PREDICATES = {
    PredicateType.PYTHAGOREAN_PREMISES: PythagoreanPremises,
    PredicateType.PYTHAGOREAN_CONCLUSIONS: PythagoreanConclusions,
}
COMPUTE = {
    PredicateType.A_COMPUTE: ACompute,
    PredicateType.R_COMPUTE: RCompute,
    PredicateType.L_COMPUTE: LCompute,
}

PREDICATES = (
    SYMBOLIC_PREDICATES | NUMERICAL_PREDICATES | INTEGRATED_PREDICATES | COMPUTE
)

NAME_TO_PREDICATE = {
    predicate_type: predicate for predicate_type, predicate in PREDICATES.items()
}


def predicate_class_from_type(
    predicate_type: PredicateType,
) -> type[PredicateInterface]:
    return NAME_TO_PREDICATE[predicate_type]  # type: ignore


Predicate: TypeAlias = Annotated[
    Circumcenter
    | Cong
    | Coll
    | NColl
    | Cyclic
    | ConstantAngle
    | ACompute
    | LCompute
    | ConstantLength
    | ConstantRatio
    | EqAngle
    | Diff
    | RCompute
    | ObtuseAngle
    | MidPoint
    | EqRatio
    | Perp
    | NPerp
    | Para
    | NPara
    | SameClock
    | SameSide
    | NSameSide
    | PythagoreanPremises
    | PythagoreanConclusions
    | SquaredConstantLength
    | SquaredConstantRatio
    | SimtriClock
    | SimtriReflect
    | ContriClock
    | ContriReflect
    | AngleEquation
    | LengthEquation,
    Field(discriminator="predicate_type"),
]


def predicate_from_construction(
    construction: PredicateConstruction, points_registry: PointsRegisty
) -> Predicate | None:
    predicate_class = predicate_class_from_type(construction.predicate_type)
    canonical_args = predicate_class.preparse(construction.args)
    if canonical_args is None:
        return None

    match construction.predicate_type:
        case PredicateType.CIRCUMCENTER:
            center, *points_around = points_registry.names2points(canonical_args)
            return Circumcenter(center=center, points=tuple(points_around))
        case PredicateType.CONGRUENT:
            a, b, c, d = points_registry.names2points(canonical_args)
            return Cong(segment1=(a, b), segment2=(c, d))
        case PredicateType.COLLINEAR:
            a, b, c = points_registry.names2points(canonical_args)
            return Coll(points=(a, b, c))
        case PredicateType.CYCLIC:
            return Cyclic(points=tuple(points_registry.names2points(canonical_args)))
        case PredicateType.CONSTANT_ANGLE:
            a, b, c, d = points_registry.names2points(canonical_args[:-1])
            y = str_to_fraction(construction.args[-1])
            return ConstantAngle(line1=(a, b), line2=(c, d), angle=y)
        case PredicateType.A_COMPUTE:
            a, b, c, d = points_registry.names2points(canonical_args)
            return ACompute(segment1=(a, b), segment2=(c, d))
        case PredicateType.N_COLLINEAR:
            return NColl(points=tuple(points_registry.names2points(canonical_args)))
        case PredicateType.CONSTANT_LENGTH:
            a, b = points_registry.names2points(canonical_args[:-1])
            length = str_to_fraction(construction.args[-1])
            return ConstantLength(segment=(a, b), length=length)
        case PredicateType.L_COMPUTE:
            a, b = points_registry.names2points(canonical_args)
            return LCompute(segment=(a, b))
        case PredicateType.CONSTANT_RATIO:
            a, b, c, d = points_registry.names2points(canonical_args[:-1])
            ratio = str_to_fraction(construction.args[-1])
            return ConstantRatio(ratio=((a, b), (c, d)), value=ratio)
        case PredicateType.R_COMPUTE:
            a, b, c, d = points_registry.names2points(canonical_args)
            return RCompute(ratio=((a, b), (c, d)))
        case PredicateType.DIFFERENT:
            a, b = points_registry.names2points(canonical_args)
            return Diff(points=(a, b))
        case PredicateType.EQUAL_ANGLES:
            a, b, c, d, e, f, g, h = points_registry.names2points(canonical_args)
            return EqAngle(angle1=((a, b), (c, d)), angle2=((e, f), (g, h)))
        case PredicateType.EQUAL_RATIOS:
            a, b, c, d, e, f, g, h = points_registry.names2points(canonical_args)
            return EqRatio(ratio1=((a, b), (c, d)), ratio2=((e, f), (g, h)))
        case PredicateType.PYTHAGOREAN_PREMISES:
            a, b, c = points_registry.names2points(canonical_args)
            return PythagoreanPremises(A=a, B=b, C=c)
        case PredicateType.PYTHAGOREAN_CONCLUSIONS:
            a, b, c = points_registry.names2points(canonical_args)
            return PythagoreanConclusions(A=a, B=b, C=c)
        case PredicateType.OBTUSE_ANGLE:
            a, b, c = points_registry.names2points(canonical_args)
            return ObtuseAngle(head1=a, corner=b, head2=c)
        case PredicateType.MIDPOINT:
            m, a, b = points_registry.names2points(canonical_args)
            return MidPoint(midpoint=m, segment=(a, b))
        case PredicateType.PERPENDICULAR:
            a, b, c, d = points_registry.names2points(canonical_args)
            return Perp(line1=(a, b), line2=(c, d))
        case PredicateType.N_PERPENDICULAR:
            a, b, c, d = points_registry.names2points(canonical_args)
            return NPerp(line1=(a, b), line2=(c, d))
        case PredicateType.PARALLEL:
            a, b, c, d = points_registry.names2points(canonical_args)
            return Para(line1=(a, b), line2=(c, d))
        case PredicateType.N_PARALLEL:
            a, b, c, d = points_registry.names2points(canonical_args)
            return NPara(line1=(a, b), line2=(c, d))
        case PredicateType.SAME_CLOCK:
            a, b, c, p, q, r = points_registry.names2points(canonical_args)
            return SameClock(triangle1=(a, b, c), triangle2=(p, q, r))
        case PredicateType.SAME_SIDE:
            a, b, c, p, q, r = points_registry.names2points(canonical_args)
            return SameSide(triangle1=(a, b, c), triangle2=(p, q, r))
        case PredicateType.N_SAME_SIDE:
            a, b, c, p, q, r = points_registry.names2points(canonical_args)
            return NSameSide(triangle1=(a, b, c), triangle2=(p, q, r))
        case PredicateType.SIMTRI_CLOCK:
            a, b, c, p, q, r = points_registry.names2points(canonical_args)
            return SimtriClock(triangle1=(a, b, c), triangle2=(p, q, r))
        case PredicateType.SIMTRI_REFLECT:
            a, b, c, p, q, r = points_registry.names2points(canonical_args)
            return SimtriReflect(triangle1=(a, b, c), triangle2=(p, q, r))
        case PredicateType.CONTRI_CLOCK:
            a, b, c, p, q, r = points_registry.names2points(canonical_args)
            return ContriClock(triangle1=(a, b, c), triangle2=(p, q, r))
        case PredicateType.CONTRI_REFLECT:
            a, b, c, p, q, r = points_registry.names2points(canonical_args)
            return ContriReflect(triangle1=(a, b, c), triangle2=(p, q, r))
        case PredicateType.SQUARED_CONSTANT_LENGTH:
            a, b = points_registry.names2points(canonical_args[:-1])
            square_length = str_to_fraction(canonical_args[-1])
            return SquaredConstantLength(segment=(a, b), square_length=square_length)
        case PredicateType.SQUARED_CONSTANT_RATIO:
            a, b, c, d = points_registry.names2points(canonical_args[:-1])
            square_ratio = str_to_fraction(canonical_args[-1])
            return SquaredConstantRatio(
                segment1=(a, b), segment2=(c, d), square_ratio=square_ratio
            )
        case PredicateType.ANGLE_EQUATION:
            raise NotImplementedError("AngleEquationPredicate is not implemented")
        case PredicateType.LENGTH_EQUATION:
            raise NotImplementedError("LengthEquationPredicate is not implemented")
    raise ValueError(f"Unknown predicate type: {construction.name}")



import logging
from typing import Literal

import numpy
from pydantic import BaseModel

from newclid.heuristics._index import HeuristicName
from newclid.heuristics._interface import Heuristic, HeuristicSetup
from newclid.heuristics.alphabet import get_available_from_alphabet
from newclid.jgex.clause import JGEXClause
from newclid.predicate_types import PredicateArgument

LOGGER = logging.getLogger(__name__)


class AngleVerticesHeuristicConfig(BaseModel):
    heuristic_name: Literal[HeuristicName.ANGLE_VERTICES] = HeuristicName.ANGLE_VERTICES


class AngleVerticesHeuristic(Heuristic):
    """
    Add all angle vertices defined at the predicate of the JGEX problem.

    Args:
        problem (JGEXProblem): The JGEX problem to which angle vertices will be added.
        angles (List): A list of initial angles defined in the JGEX problem.

    Returns:
        JGEXProblem: The modified JGEX problem with added angle vertices.
    """

    def new_clauses(
        self,
        setup: HeuristicSetup,
        max_new_points: int,
        rng: numpy.random.Generator,
    ) -> list[JGEXClause]:
        angles = list(setup.angles)

        added_points: list[PredicateArgument] = []
        new_clauses: list[JGEXClause] = []
        for _index in range(max_new_points):
            if not angles:
                break

            angle_index = rng.choice(len(angles))
            angle = angles.pop(angle_index)

            point_1 = angle.line_1[0]
            point_2 = angle.line_1[1]
            point_3 = angle.line_2[0]
            point_4 = angle.line_2[1]

            new_point = get_available_from_alphabet(list(setup.points) + added_points)
            if new_point is None:
                LOGGER.warning(
                    "No available points left in the alphabet, skipping angle vertex addition."
                )
                break

            added_points.append(new_point)
            new_clause_string = f"{new_point} = on_line {new_point} {point_1} {point_2}, on_line {new_point} {point_3} {point_4}"
            new_clause = JGEXClause.from_str(new_clause_string)[0]
            new_clauses.append(new_clause)

        return new_clauses



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Literal

from newclid.numerical import close_enough
from newclid.predicate_types import PredicateArgument
from newclid.predicates._index import PredicateType
from newclid.predicates._interface import PredicateInterface
from newclid.predicates.constant_length import ConstantLength
from newclid.predicates.perpendicularity import Perp
from newclid.symbols.points_registry import Point
from newclid.tools import InfQuotientError, get_quotient

if TYPE_CHECKING:
    from newclid.predicates import Predicate
    from newclid.proof_state import ProofState


LOGGER = logging.getLogger(__name__)


class PythagoreanPremises(PredicateInterface):
    """PythagoreanPremises a b c
    abc is in the form of a right angled triangle. ab is perpendicular to ac
    """

    predicate_type: Literal[PredicateType.PYTHAGOREAN_PREMISES] = (
        PredicateType.PYTHAGOREAN_PREMISES
    )
    A: Point
    B: Point
    C: Point

    @staticmethod
    def preparse(
        args: tuple[PredicateArgument, ...],
    ) -> tuple[PredicateArgument, ...] | None:
        a, b, c = args
        if a == b or a == c:
            return None
        b, c = sorted((b, c))
        return (a, b, c)

    def check_numerical(self) -> bool:
        return close_enough(
            abs((self.A.num - self.B.num).dot(self.A.num - self.C.num)), 0
        )

    def check(self, proof_state: ProofState) -> bool | None:
        return self.why(proof_state) is not None

    def why(self, proof_state: ProofState) -> tuple[str, tuple[Predicate, ...]] | None:
        a, b, c = self.A, self.B, self.C
        perp = Perp(line1=(a, b), line2=(a, c))
        perp_check = proof_state.check(perp)
        try:
            ab = ConstantLength(
                segment=(a, b), length=get_quotient(a.num.distance(b.num))
            )
            ac = ConstantLength(
                segment=(a, c), length=get_quotient(a.num.distance(c.num))
            )
            bc = ConstantLength(
                segment=(b, c), length=get_quotient(b.num.distance(c.num))
            )
        except InfQuotientError:
            return None

        check_ab = proof_state.check(ab)
        check_ac = proof_state.check(ac)
        check_bc = proof_state.check(bc)

        justification: tuple[ConstantLength | Perp, ...] | None = None
        if check_ab and check_ac and check_bc:
            justification = (ab, ac, bc)
        if perp_check and check_ac and check_bc:
            justification = (perp, ac, bc)
        if perp_check and check_ab and check_bc:
            justification = (ab, perp, bc)
        if perp_check and check_ab and check_ac:
            justification = (ab, ac, perp)
        if justification is None:
            return None
        return ("Pythagorean verification", justification)

    def to_tokens(self) -> tuple[PredicateArgument, ...]:
        return tuple(PredicateArgument(p.name) for p in (self.A, self.B, self.C))

    def __str__(self) -> str:
        return f"Pythagorean Theorem's premises on {self.A}, {self.B}, {self.C} are satisfied"


class PythagoreanConclusions(PredicateInterface):
    predicate_type: Literal[PredicateType.PYTHAGOREAN_CONCLUSIONS] = (
        PredicateType.PYTHAGOREAN_CONCLUSIONS
    )
    A: Point
    B: Point
    C: Point

    @staticmethod
    def preparse(
        args: tuple[PredicateArgument, ...],
    ) -> tuple[PredicateArgument, ...] | None:
        return PythagoreanPremises.preparse(args)

    def check_numerical(self) -> bool:
        return PythagoreanPremises(A=self.A, B=self.B, C=self.C).check_numerical()

    def add(self, proof_state: ProofState) -> tuple[PredicateInterface, ...]:
        a, b, c = self.A, self.B, self.C
        perp = Perp(line1=(a, b), line2=(a, c))
        perp_check = proof_state.check(perp)
        consequences: list[PredicateInterface] = []
        if not perp_check:
            consequences.append(perp)
        try:
            ab = ConstantLength(
                segment=(a, b), length=get_quotient(a.num.distance(b.num))
            )
            ac = ConstantLength(
                segment=(a, c), length=get_quotient(a.num.distance(c.num))
            )
            bc = ConstantLength(
                segment=(b, c), length=get_quotient(b.num.distance(c.num))
            )
        except InfQuotientError:
            LOGGER.info(
                "lconst result could be added, but the irrational number len cannot be represented."
            )
            return ()

        check_ab = proof_state.check(ab)
        check_ac = proof_state.check(ac)
        check_bc = proof_state.check(bc)

        if not check_ab:
            consequences.append(ab)
        if not check_ac:
            consequences.append(ac)
        if not check_bc:
            consequences.append(bc)
        return tuple(consequences)

    def to_tokens(self) -> tuple[PredicateArgument, ...]:
        return tuple(PredicateArgument(p.name) for p in (self.A, self.B, self.C))

    def __str__(self) -> str:
        return f"Pythagorean Theorem's conclusion on {self.A}, {self.B}, {self.C} is satisfied"



import numpy as np
import pytest
from newclid.api import GeometricSolverBuilder
from newclid.jgex.problem_builder import JGEXProblemBuilder


class TestConstants:
    @pytest.fixture(autouse=True)
    def setUpClass(self):
        self.rng = np.random.default_rng(233)
        self.solver_builder = GeometricSolverBuilder(rng=self.rng)
        self.problem_builder = JGEXProblemBuilder(rng=self.rng)

    def test_aconst_deg(self):
        """Should be able to prescribe and check a constant angle in degree"""
        solver = self.solver_builder.build(
            self.problem_builder.with_problem_from_txt(
                "a b = segment a b; c = free c; x = aconst a b c x 63o; y = aconst a b c y 153o ? aconst c x c y 90o",
            ).build()
        )
        success = solver.run()
        assert success

    @pytest.mark.skip(reason="acompute is not supported by Yuclid")
    def test_acompute(self):
        solver = self.solver_builder.build(
            self.problem_builder.with_problem_from_txt(
                "a b = segment a b; c = free c; x = aconst a b c x 63o; y = aconst a b c y 153o ? acompute c x c y",
            ).build()
        )
        success = solver.run()
        assert success

    def test_aconst_pi_frac(self):
        """Should be able to prescribe and check a constant angle as pi fraction"""
        solver = self.solver_builder.build(
            self.problem_builder.with_problem_from_txt(
                "a b = segment a b; "
                "c = free c; "
                "x = aconst a b c x 7pi/20; "
                "y = aconst a b c y 17pi/20 "
                "? aconst c x c y 1pi/2"
            ).build()
        )
        success = solver.run()
        assert success

    def test_s_angle_deg(self):
        """Should be able to prescribe and check a constant s_angle in degree"""
        solver = self.solver_builder.build(
            self.problem_builder.with_problem_from_txt(
                "a b = segment a b; x = s_angle a b x 63o; y = s_angle a b y 153o ? aconst x b b y 90o"
            ).build()
        )
        success = solver.run()
        assert success

    def test_s_angle_deg_not_perp(self):
        """Should be able to prescribe and check a constant s_angle in degree"""
        solver = self.solver_builder.build(
            self.problem_builder.with_problem_from_txt(
                "a b = segment a b; x = s_angle a b x 63o; y = s_angle a b y 143o ? aconst x b b y 80o",
            ).build()
        )
        success = solver.run()
        assert success

    def test_s_angle_pi_frac(self):
        """Should be able to prescribe and check a constant s_angle as pi fraction"""
        solver = self.solver_builder.build(
            self.problem_builder.with_problem_from_txt(
                "a b = segment a b; x = s_angle a b x 7pi/20; y = s_angle a b y 17pi/20 ? aconst x b b y 1pi/2",
            ).build()
        )
        success = solver.run()
        assert success

    def test_s_angle_in_perp_out(self):
        """Should be able to get a perp from prescribed s_angle in degree"""
        solver = self.solver_builder.build(
            self.problem_builder.with_problem_from_txt(
                "a b = segment a b; x = s_angle a b x 63o; y = s_angle b a y 153o ? perp b x a y",
            ).build()
        )
        success = solver.run()
        assert success

    def test_s_angle_in_aconst_out(self):
        """Should be able to check aconst in radiant
        from s_angle presciption in degrees"""
        solver = self.solver_builder.build(
            self.problem_builder.with_problem_from_txt(
                "a b = segment a b; x = s_angle a b x 63o; y = s_angle b a y 153o ? aconst b x a y 1pi/2"
            ).build()
        )
        success = solver.run()
        assert success

    def test_rconst(self):
        """Shoulb be able to prescribe and check a constant ratio"""
        solver = self.solver_builder.build(
            self.problem_builder.with_problem_from_txt(
                "a b = segment a b; c = free c; d = rconst a b c d 3/4 ? rconst a b c d 3/4"
            ).build()
        )
        success = solver.run()
        assert success

    def test_rconst_as_theorem_conclusion(self):
        solver = self.solver_builder.build(
            self.problem_builder.with_problem_from_txt(
                "a b = segment a b; m = midpoint m a b ? rconst m a a b 1/2",
            ).build()
        )
        success = solver.run()
        assert success

    @pytest.mark.skip
    def test_rcompute(self):
        solver = self.solver_builder.build(
            self.problem_builder.with_problem_from_txt(
                "a b = segment a b; m = midpoint m a b ? rcompute m a a b",
            ).build()
        )
        success = solver.run()
        assert success

    def test_triangle12_in_rconst_out(self):
        """Should obtain a constant ratio from triangle12"""
        solver = self.solver_builder.build(
            self.problem_builder.with_problem_from_txt(
                "a b c = triangle12 a b c ? rconst a b a c 1/2"
            ).build()
        )
        success = solver.run()
        assert success

    def test_lconst(self):
        """Should be able to prescribe a constant lenght"""
        solver = self.solver_builder.build(
            self.problem_builder.with_problem_from_txt(
                "a = free a; b = lconst b a 3 ? lconst b a 3"
            ).build()
        )
        success = solver.run()
        assert success

    @pytest.mark.skip(reason="lcompute is not supported by Yuclid")
    def test_lcompute(self):
        """Should be able to prescribe a constant lenght"""
        solver = self.solver_builder.build(
            self.problem_builder.with_problem_from_txt(
                "a = free a; b = lconst b a 3 ? lcompute a b"
            ).build()
        )
        success = solver.run()
        assert success



from dataclasses import dataclass

from matplotlib.artist import Artist
from matplotlib.axes import Axes
from matplotlib.typing import ColorType

from newclid.animation.artists_in_animation import ArtistInAnimation, PointArtists
from newclid.draw.predicates import draw_predicate
from newclid.draw.theme import DrawTheme
from newclid.predicate_types import PredicateArgument
from newclid.predicates import NUMERICAL_PREDICATES, Predicate
from newclid.problem import predicate_points
from newclid.proof_data import PredicateInProof
from newclid.symbols.symbols_registry import SymbolsRegistry


@dataclass
class GoalArtists:
    goal: Predicate
    artists: list[ArtistInAnimation]
    used_points: list[PointArtists]

    def highlight(
        self,
        aux_point_names: set[PredicateArgument],
        highlight_color: ColorType | None = None,
    ) -> set[Artist]:
        updated_artists: set[Artist] = set()
        for artist in self.artists:
            if artist.highlight(highlight_color):
                updated_artists.add(artist.artist)
        for point in self.used_points:
            color = highlight_color if point.name not in aux_point_names else None
            updated_artists.update(point.highlight(color=color))
        return updated_artists

    def hide(self) -> set[Artist]:
        updated_artists: set[Artist] = set()
        for artist in self.artists:
            if artist.hide():
                updated_artists.add(artist.artist)
        return updated_artists


def init_goals_to_draw(
    ax: Axes,
    proven_goals: list[PredicateInProof],
    point_name_to_artist: dict[str, PointArtists],
    symbols: SymbolsRegistry,
    theme: DrawTheme,
) -> dict[str, GoalArtists]:
    goals_artists: dict[str, GoalArtists] = {}
    for goal in proven_goals:
        if goal.predicate.predicate_type in NUMERICAL_PREDICATES:
            continue

        used_points: list[PointArtists] = []
        for point in predicate_points(goal.predicate):
            used_points.append(point_name_to_artist[point])

        artists = draw_predicate(ax, goal.predicate, symbols, theme=theme)
        goals_artists[goal.id] = GoalArtists(
            goal=goal.predicate,
            artists=[ArtistInAnimation(artist) for artist in artists],
            used_points=used_points,
        )
    return goals_artists



import itertools
from typing import Iterator, TypeVar

P = TypeVar("P")
V = TypeVar("V")


def generate_permutations_as_dicts(
    points: list[P], variables: list[V]
) -> Iterator[dict[V, P]]:
    """Generates dictionaries from permutations of 'points' mapped to 'variables'.

    This function is a Python implementation of the C++ `generate_permutations_as_dicts`.
    It behaves like Python's itertools.permutations, where each permutation
    of points is zipped with variables to form a dictionary.

    Args:
        points: A list of strings to be permuted.
        variables: A list of strings to be used as keys. The length of
                   permutations will be equal to len(variables).

    Yields:
        A dictionary where keys are from 'variables' and values are
        from a permutation of 'points'.
    """
    r = len(variables)
    for p in itertools.permutations(points, r):
        yield dict(zip(variables, p))



import io
import sys

import numpy as np
import pytest
from newclid.agent.human_agent import HumanAgent
from newclid.api import GeometricSolverBuilder
from newclid.jgex.problem_builder import JGEXProblemBuilder


class TestHumanAgent:
    @pytest.fixture(autouse=True)
    def setUpClass(self):
        self.rng = np.random.default_rng(998244353)
        self.problem_builder = JGEXProblemBuilder(rng=self.rng)
        self.solver_builder = GeometricSolverBuilder(rng=self.rng).with_deductive_agent(
            HumanAgent()
        )

    @pytest.mark.skip(reason="Annoying popup for unnecessary feature")
    def test_graphics(self):
        sys.stdin = io.StringIO("1\n5\nall\n6\n")
        self.problem_builder.with_problem_from_txt(
            "a b c = triangle a b c;h = on_tline h b a c, on_tline h c a b ? perp a h b c"
        )
        solver = self.solver_builder.with_rules([]).build(self.problem_builder)
        solver.run()

    @pytest.mark.skip(reason="Annoying popup for unnecessary feature")
    def test_add_construction(self):
        sys.stdin = io.StringIO("2\nd = on_line d a c, on_line d b h\n3\n6\n")
        self.problem_builder.with_problem_from_txt(
            "a b c = triangle a b c;h = on_tline h b a c, on_tline h c a b ? perp a h b c"
        )
        solver = self.solver_builder.with_rules([]).build(self.problem_builder)
        assert solver.run()



import itertools
from typing import Iterable

from pydantic import BaseModel

from newclid.jgex.constructions.free import FREE_CONSTRUCTIONS
from newclid.jgex.formulation import JGEXFormulation
from newclid.predicate_types import PredicateArgument
from newclid.predicates._index import PredicateType
from newclid.problem import PredicateConstruction

free_names = [definition.name for definition in FREE_CONSTRUCTIONS]


class LineHeuristic(BaseModel):
    points: tuple[PredicateArgument, ...]


class CircleHeuristic(BaseModel):
    points: tuple[PredicateArgument, ...]


class AngleHeuristic(BaseModel):
    line_1: tuple[PredicateArgument, PredicateArgument]
    line_2: tuple[PredicateArgument, PredicateArgument]


def read_geometric_objects_from_predicate_constructions(
    problem: JGEXFormulation,
    predicate_constructions: Iterable[PredicateConstruction],
) -> tuple[
    list[LineHeuristic],
    list[CircleHeuristic],
    list[AngleHeuristic],
    list[PredicateArgument],
]:
    """Read geometric objects from predicate constructions.

    Args:
        predicate_constructions (Iterable[StatementConstruction]): The predicate constructions to read geometric objects from.

    Returns:
        lines, circles, angles, free_points: The geometric objects read from the predicate constructions.
    """

    # Not working as information is not been stored in the symbols_registry by yuclid,
    # lines = solver.proof_state.symbols.lines.nodes_of_type(Line)
    # circles = solver.proof_state.symbols.circles.nodes_of_type(Circle)
    lines: list[LineHeuristic] = []
    circles: list[CircleHeuristic] = []
    angles: list[AngleHeuristic] = []
    free_points: list[PredicateArgument] = []

    # Gathering the lines involving free constructions
    for construction in problem.clauses:
        if construction.constructions[0].name in free_names:
            for point in construction.points:
                point_as_argument = PredicateArgument(point)
                free_points.append(point_as_argument)

    for p1, p2 in itertools.combinations(free_points, 2):
        lines.append(LineHeuristic(points=(p1, p2)))

    for stmt in predicate_constructions:
        match stmt.predicate_type:
            case PredicateType.CYCLIC:
                circles.append(CircleHeuristic(points=stmt.args))
            case PredicateType.CIRCUMCENTER:
                circles.append(CircleHeuristic(points=stmt.args))
            case PredicateType.COLLINEAR:
                lines.append(LineHeuristic(points=stmt.args))
            case PredicateType.PERPENDICULAR:
                lines.append(LineHeuristic(points=(stmt.args[0], stmt.args[1])))
                lines.append(LineHeuristic(points=(stmt.args[2], stmt.args[3])))
                angles.append(
                    AngleHeuristic(
                        line_1=(stmt.args[0], stmt.args[1]),
                        line_2=(stmt.args[2], stmt.args[3]),
                    )
                )
            case PredicateType.PARALLEL:
                lines.append(LineHeuristic(points=(stmt.args[0], stmt.args[1])))
                lines.append(LineHeuristic(points=(stmt.args[2], stmt.args[3])))
            case PredicateType.EQUAL_ANGLES:
                angles.append(
                    AngleHeuristic(
                        line_1=(stmt.args[0], stmt.args[1]),
                        line_2=(stmt.args[2], stmt.args[3]),
                    )
                )
                angles.append(
                    AngleHeuristic(
                        line_1=(stmt.args[4], stmt.args[5]),
                        line_2=(stmt.args[6], stmt.args[7]),
                    )
                )
            case _:
                pass

    merged_lines: list[LineHeuristic] = []
    skipped_lines_indexes: set[int] = set()
    for i, line1 in enumerate(lines):
        if i in skipped_lines_indexes:
            continue
        for j in range(i, len(lines)):
            have_two_points_in_common = (
                len(set(line1.points) & set(lines[j].points)) > 1
            )
            if have_two_points_in_common:
                merged_line = tuple(set(line1.points) | set(lines[j].points))
                skipped_lines_indexes.add(j)
                merged_lines.append(LineHeuristic(points=merged_line))

    merged_circles: list[CircleHeuristic] = []
    skipped_circles_indexes: set[int] = set()
    for i, circle1 in enumerate(circles):
        if i in skipped_circles_indexes:
            continue
        for j in range(i, len(circles)):
            have_three_points_in_common = (
                len(set(circle1.points) & set(circles[j].points)) > 2
            )
            if have_three_points_in_common:
                merged_circle = tuple(set(circle1.points) | set(circles[j].points))
                skipped_circles_indexes.add(j)
                merged_circles.append(CircleHeuristic(points=merged_circle))

    return merged_lines, merged_circles, angles, free_points



"""Implements the proof state."""

from __future__ import annotations

import logging

from numpy.random import Generator as RngGenerator

from newclid.deductors.deductor_interface import (
    Deductor,
    add_to_deductors,
    check_from_deductors,
    justify_from_deductors,
)
from newclid.justifications.justification import (
    Assumption,
    Justification,
    NumericalCheck,
)
from newclid.justifications.justify_predicates import justify_predicate
from newclid.justifications.predicates_graph import PredicatesGraph
from newclid.predicates import (
    NUMERICAL_PREDICATES,
    Predicate,
    predicate_from_construction,
)
from newclid.problem import ProblemSetup
from newclid.rng import setup_rng
from newclid.rule import Rule
from newclid.rule_matching.interface import RuleMatcher
from newclid.symbols.symbols_registry import SymbolsRegistry

LOGGER = logging.getLogger(__name__)


class ProofBuildError(Exception):
    pass


class ProofState:
    """Object representing the proof state."""

    def __init__(
        self,
        *,
        problem: ProblemSetup,
        rule_matcher: RuleMatcher,
        deductors: list[Deductor],
        rng: RngGenerator | int | None = None,
    ):
        self.problem = problem
        self._matcher = rule_matcher
        self.deductors = deductors
        self._numerical_check_cache: dict[Predicate, bool] = {}
        self.graph = PredicatesGraph()
        self.symbols = SymbolsRegistry()
        self.rng = setup_rng(rng)
        self._BLACKLIST: set[Justification] = set()

        self._add_points_to_symbols_graph()
        self._add_assumptions_to_pred_graph()
        self.goals = self._goals_from_problem()

    def match_theorem(self, rule: Rule) -> set[Justification]:
        """Match a rule to the proof state.

        Returns:
            A set of dependencies that would be added to the proof state if the rule was applied.
        """
        return self._matcher.match_theorem(rule, proof=self)

    def apply(self, justification: Justification) -> bool:
        """Add the justification to the proof dependency graph.

        Returns:
            True if the predicate is a new one, false otherwise.
        """
        if justification in self._BLACKLIST:
            return False

        if not self.check_numerical(justification.predicate):
            LOGGER.warning(
                f"Trying to add justification {justification} the conclusion of which is numerically false"
            )
            self._BLACKLIST.add(justification)
            return False

        dep_for_predicate = self.graph.hyper_graph.get(justification.predicate)
        already_proven = dep_for_predicate is not None
        if already_proven:
            return False

        self.graph.hyper_graph[justification.predicate] = justification
        justification.predicate.add(self)
        _symbols = justification.predicate.symbols(self.symbols)
        add_to_deductors(justification, self.deductors, self.symbols)
        return True

    def check(self, predicate: Predicate) -> bool:
        """Symbolically check if the predicate is currently considered True."""
        if predicate in self.graph.hyper_graph:
            return True
        if not predicate.check_numerical():
            return False
        symbolic_check = predicate.check(self)
        if symbolic_check:
            self.justify(predicate)
            return True
        if symbolic_check is None:
            deductors_check = check_from_deductors(
                predicate, self.deductors, self.symbols
            )
            if deductors_check:
                return True
        return False

    def check_numerical(self, predicate: Predicate) -> bool:
        """Check if the predicate is numerically sound."""
        if predicate in self._numerical_check_cache:
            return self._numerical_check_cache[predicate]
        res = predicate.check_numerical()
        self._numerical_check_cache[predicate] = res
        return res

    def justify(self, predicate: Predicate) -> Justification:
        justification_cached = self.graph.hyper_graph.get(predicate)
        if justification_cached is not None:
            return justification_cached

        justification = justify_predicate(predicate, self.deductors, self.symbols)
        if justification is None:
            justification = justify_from_deductors(
                predicate, self.deductors, self.symbols
            )
        if justification is None:
            raise ValueError(f"Could not justify predicate {predicate}")

        self.apply(justification)
        return justification

    def check_goals(self) -> bool:
        """Check if the goals are all symbolically checked."""
        if not self.goals:
            return False
        for goal in self.goals:
            if not self.check(goal):
                return False
        return True

    def _add_points_to_symbols_graph(self) -> None:
        for point in self.problem.points:
            self.symbols.points.add_point(point)

    def _add_assumptions_to_pred_graph(self) -> None:
        for predicate_construction in self.problem.assumptions:
            predicate = predicate_from_construction(
                predicate_construction,
                points_registry=self.symbols.points,
            )
            if predicate is None:
                raise ProofBuildError(
                    f"Assumption '{predicate_construction}' cannot be built."
                )
            if not self.check_numerical(predicate):
                raise ProofBuildError(
                    f"Assumption '{predicate_construction}' is numerically false."
                )
            dependency: Justification
            if predicate.predicate_type in NUMERICAL_PREDICATES:
                dependency = NumericalCheck(predicate=predicate)
            else:
                dependency = Assumption(predicate=predicate)
            self.apply(dependency)

    def _goals_from_problem(self) -> list[Predicate]:
        goals: list[Predicate] = []
        for goal in self.problem.goals:
            goal_predicate = predicate_from_construction(
                goal, points_registry=self.symbols.points
            )
            if goal_predicate is None:
                continue
            if not self.check_numerical(goal_predicate):
                raise ProofBuildError(f"Goal {goal} is numerically false.")
            goals.append(goal_predicate)
        return goals



from __future__ import annotations

from typing import Literal

from newclid.numerical.check import same_clock
from newclid.predicate_types import PredicateArgument
from newclid.predicates._index import PredicateType
from newclid.predicates._interface import NumericalPredicate
from newclid.symbols.points_registry import Triangle


class SameClock(NumericalPredicate):
    """sameclock a b c x y z -"""

    predicate_type: Literal[PredicateType.SAME_CLOCK] = PredicateType.SAME_CLOCK
    triangle1: Triangle
    triangle2: Triangle

    @staticmethod
    def preparse(
        args: tuple[PredicateArgument, ...],
    ) -> tuple[PredicateArgument, ...] | None:
        a, b, c, x, y, z = args
        if a == b or b == c or a == c or x == y or y == z or x == z:
            return None
        group = min((a, b, c), (b, c, a), (c, a, b))
        groupr = min((a, c, b), (c, b, a), (b, a, c))
        group1 = min((x, y, z), (y, z, x), (z, x, y))
        group1r = min((x, z, y), (z, y, x), (y, x, z))
        return min(group + group1, group1 + group, groupr + group1r, group1r + groupr)

    def check_numerical(self) -> bool:
        a, b, c = self.triangle1
        x, y, z = self.triangle2
        return same_clock(a.num, b.num, c.num, x.num, y.num, z.num)

    def to_tokens(self) -> tuple[PredicateArgument, ...]:
        return tuple(PredicateArgument(p.name) for p in self.triangle1 + self.triangle2)

    def __str__(self) -> str:
        a, b, c = self.triangle1
        x, y, z = self.triangle2
        return f"▲{a}{b}{c} has the same orientation as ▲{x}{y}{z}"



import numpy as np
import pytest
from newclid.all_rules import R43_ORTHOCENTER_THEOREM
from newclid.api import GeometricSolverBuilder
from newclid.jgex.clause import JGEXClause, JGEXConstruction
from newclid.jgex.jgex_setup_data import (
    JGEXClauseInProof,
    JGEXSetupData,
    JGEXSetupId,
    PredicateConstructionInSetup,
    jgex_clauses_to_setup_data,
)
from newclid.jgex.problem_builder import JGEXProblemBuilder
from newclid.predicate_types import PredicateArgument
from newclid.problem import (
    PredicateConstruction,
    predicate_from_str,
    predicate_to_construction,
)
from newclid.proof_data import (
    PredicateInProof,
    ProofData,
    ProofId,
    ProofStep,
    proof_data_from_state,
)
from newclid.rule import RuleApplication
from newclid.tools import pretty_basemodel_diff


def p(s: str) -> PredicateArgument:
    return PredicateArgument(s)


class TestJGEXProofData:
    def test_setup_data_from_sampled_datagen_problem(self):
        """Test that this sample out of datagen has correct proof data."""
        rng = np.random.default_rng(42)
        jgex_problem_builder = (
            JGEXProblemBuilder(rng=rng)
            .include_auxiliary_clauses()
            .with_problem_from_txt(
                "a b c d = isquare a b c d; e f g h = centroid e f g h d c a; i = on_line i a h; j = eqdistance j c h d; k = parallelogram k g i b; l = intersection_lp l d j k g a; m = intersection_lc m l e d; n o = trisegment n o k j; p q r s = cc_tangent p q r s d m c a; t u v w = centroid t u v w r p e; x = on_bline x c d, on_tline x l e h ? eqangle c q g q p s q s"
            )
        )
        solver = GeometricSolverBuilder(rng=rng).build(jgex_problem_builder.build())
        success = solver.run()
        assert success
        assert jgex_problem_builder.jgex_problem is not None
        setup_data, _predicates_ids = jgex_clauses_to_setup_data(
            setup_clauses=list(jgex_problem_builder.jgex_problem.setup_clauses),
            aux_clauses=list(jgex_problem_builder.jgex_problem.auxiliary_clauses),
            goals=[
                predicate_to_construction(goal) for goal in solver.proof_state.goals
            ],
            clauses_consequences=jgex_problem_builder.clauses_consequences,
        )
        expected_square_clauses_in_proof = JGEXClauseInProof(
            added_points=("a", "b", "c", "d"),
            constructions=(JGEXConstruction.from_str("isquare a b c d"),),
            predicates=(
                PredicateConstructionInSetup(
                    id=JGEXSetupId("C0"),
                    construction=PredicateConstruction.from_str("perp a b b c"),
                ),
                PredicateConstructionInSetup(
                    id=JGEXSetupId("C1"),
                    construction=PredicateConstruction.from_str("cong a b b c"),
                ),
                PredicateConstructionInSetup(
                    id=JGEXSetupId("C2"),
                    construction=PredicateConstruction.from_str("para a b c d"),
                ),
                PredicateConstructionInSetup(
                    id=JGEXSetupId("C3"),
                    construction=PredicateConstruction.from_str("para b c a d"),
                ),
                PredicateConstructionInSetup(
                    id=JGEXSetupId("C4"),
                    construction=PredicateConstruction.from_str("perp a d c d"),
                ),
                PredicateConstructionInSetup(
                    id=JGEXSetupId("C5"),
                    construction=PredicateConstruction.from_str("cong b c c d"),
                ),
                PredicateConstructionInSetup(
                    id=JGEXSetupId("C6"),
                    construction=PredicateConstruction.from_str("cong c d a d"),
                ),
                PredicateConstructionInSetup(
                    id=JGEXSetupId("C7"),
                    construction=PredicateConstruction.from_str("perp a c b d"),
                ),
                PredicateConstructionInSetup(
                    id=JGEXSetupId("C8"),
                    construction=PredicateConstruction.from_str("cong b d a c"),
                ),
            ),
        )
        actual_square_clauses_in_proof = setup_data.setup_clauses[0]
        assert actual_square_clauses_in_proof == expected_square_clauses_in_proof, (
            pretty_basemodel_diff(
                actual_square_clauses_in_proof, expected_square_clauses_in_proof
            )
        )

    def test_setup_data_from_orthocenter(self):
        """At initialisation, the proof data should be filled by construction predicates."""
        rng = np.random.default_rng(42)
        jgex_problem_builder = JGEXProblemBuilder(rng=rng).with_problem_from_txt(
            "a b c = triangle a b c; h = on_tline h b a c, on_tline h c a b ? perp a h b c"
        )
        solver = GeometricSolverBuilder(rng=rng).build(jgex_problem_builder.build())
        assert jgex_problem_builder.jgex_problem is not None
        setup_data, _predicates_ids = jgex_clauses_to_setup_data(
            setup_clauses=list(jgex_problem_builder.jgex_problem.setup_clauses),
            aux_clauses=[],
            clauses_consequences=jgex_problem_builder.clauses_consequences,
            goals=[
                predicate_to_construction(goal) for goal in solver.proof_state.goals
            ],
        )
        assert setup_data == self.orthocenter_setup_data, pretty_basemodel_diff(
            setup_data, self.orthocenter_setup_data
        )

    def test_proof_data_from_orthocenter_solved(self):
        """At initialisation, the proof data should be filled by construction predicates."""
        rng = np.random.default_rng(42)
        jgex_problem_builder = JGEXProblemBuilder(rng=rng).with_problem_from_txt(
            "a b c = triangle a b c; h = on_tline h b a c, on_tline h c a b ? perp a h b c"
        )
        solver = GeometricSolverBuilder(rng=rng).build(jgex_problem_builder.build())
        solver.run()
        assert jgex_problem_builder.jgex_problem is not None

        points = solver.proof_state.symbols.points

        proof_data = proof_data_from_state(
            goals_constructions=[
                predicate_to_construction(goal) for goal in solver.proof_state.goals
            ],
            proof_state=solver.proof_state,
        )

        goal_0 = PredicateInProof(
            id=ProofId("0"),
            predicate=predicate_from_str("perp a h b c", points),
        )

        expected_proof_data = ProofData(
            points=list(solver.proof_state.problem.points),
            proven_goals=[goal_0],
            unproven_goals=[],
            construction_assumptions=[
                PredicateInProof(
                    id=ProofId("C0"),
                    predicate=predicate_from_str("perp a b c h", points),
                ),
                PredicateInProof(
                    id=ProofId("C1"),
                    predicate=predicate_from_str("perp a c b h", points),
                ),
            ],
            numerical_checks=[],
            trivial_predicates=[],
            proof_steps=[
                ProofStep(
                    proven_predicate=goal_0,
                    justification=RuleApplication(
                        rule=R43_ORTHOCENTER_THEOREM,
                        predicate=predicate_from_str("perp a h b c", points),
                        premises=(
                            predicate_from_str("perp a b c h", points),
                            predicate_from_str("perp a c b h", points),
                        ),
                    ),
                    applied_on_predicates=(ProofId("C0"), ProofId("C1")),
                ),
            ],
        )

        assert proof_data == expected_proof_data, pretty_basemodel_diff(
            proof_data, expected_proof_data
        )

    @pytest.fixture(autouse=True)
    def setup(self):
        triangle_clause = JGEXClause(
            points=(p("a"), p("b"), p("c")),
            constructions=(JGEXConstruction.from_str("triangle a b c"),),
        )
        h_on_tlines_clause = JGEXClause(
            points=(p("h"), p("b"), p("a"), p("c")),
            constructions=(
                JGEXConstruction.from_str("on_tline h b a c"),
                JGEXConstruction.from_str("on_tline h c a b"),
            ),
        )

        self.orthocenter_setup_data = JGEXSetupData(
            goals=[
                PredicateConstructionInSetup(
                    id=JGEXSetupId("G0"),
                    construction=PredicateConstruction.from_str("perp a h b c"),
                ),
            ],
            setup_clauses=[
                JGEXClauseInProof(
                    constructions=triangle_clause.constructions,
                    predicates=(),
                    added_points=("a", "b", "c"),
                ),
                JGEXClauseInProof(
                    constructions=h_on_tlines_clause.constructions,
                    predicates=(
                        PredicateConstructionInSetup(
                            id=JGEXSetupId("C0"),
                            construction=PredicateConstruction.from_str("perp a c b h"),
                        ),
                        PredicateConstructionInSetup(
                            id=JGEXSetupId("C1"),
                            construction=PredicateConstruction.from_str("perp a b c h"),
                        ),
                    ),
                    added_points=("h",),
                ),
            ],
            aux_clauses=[],
        )



from enum import Enum
from typing import Annotated, Literal

from pydantic import BaseModel, Field

from newclid.ggb.read_elements import GGBConic, GGBLine, GGBPoint


class RelationShipType(Enum):
    Parallel = "parallel"
    Orthogonal = "orthogonal"
    BisectorLine = "bisector_line"
    BisectorPoint = "bisector_point"
    Tangent = "tangent"
    Congruent = "congruent"
    Cyclic = "cyclic"
    Collinear = "collinear"


class ParaRelationship(BaseModel):
    line: GGBLine
    parallel_to: GGBLine
    relationship_type: Literal[RelationShipType.Parallel] = RelationShipType.Parallel


class OrthoRelationship(BaseModel):
    line: GGBLine
    orthogonal_to: GGBLine
    relationship_type: Literal[RelationShipType.Orthogonal] = (
        RelationShipType.Orthogonal
    )


class BisectorLineRelationship(BaseModel):
    line: GGBLine
    bisector_1: GGBLine
    bisector_2: GGBLine
    relationship_type: Literal[RelationShipType.BisectorLine] = (
        RelationShipType.BisectorLine
    )


class BisectorPointRelationship(BaseModel):
    line: GGBLine
    angle_a: GGBPoint
    angle_vertex: GGBPoint
    angle_b: GGBPoint
    relationship_type: Literal[RelationShipType.BisectorPoint] = (
        RelationShipType.BisectorPoint
    )


class TangentRelationship(BaseModel):
    line: GGBLine
    tangent_to: GGBConic
    relationship_type: Literal[RelationShipType.Tangent] = RelationShipType.Tangent


ElementsRelationships = Annotated[
    ParaRelationship
    | OrthoRelationship
    | TangentRelationship
    | BisectorLineRelationship
    | BisectorPointRelationship,
    Field(discriminator="relationship_type"),
]



from random import random
from typing import Any, cast

import numpy as np
from matplotlib import patches
from matplotlib.artist import Artist
from matplotlib.axes import Axes
from matplotlib.lines import AxLine, Line2D

from newclid.draw.geometries import draw_circle_symbol
from newclid.draw.theme import PALETTE, DrawTheme, fill_missing
from newclid.jgex.geometries import LENGTH_UNIT
from newclid.numerical.geometries import PointNum, line_line_intersection
from newclid.predicates import Predicate
from newclid.predicates._index import PredicateType
from newclid.predicates.congruence import Cong
from newclid.predicates.equal_angles import EqAngle
from newclid.predicates.equal_ratios import EqRatio
from newclid.predicates.perpendicularity import Perp
from newclid.symbols.lines_registry import LineSymbol
from newclid.symbols.points_registry import Line, Point
from newclid.symbols.symbols_registry import SymbolsRegistry


def draw_predicate(
    ax: Axes,
    predicate: Predicate,
    symbols: SymbolsRegistry,
    theme: DrawTheme,
) -> set[Artist]:
    match predicate.predicate_type:
        case PredicateType.COLLINEAR:
            line = symbols.lines.line_containing(set(predicate.points))
            if line is None:
                raise ValueError(f"Line not found for args: {predicate.points}")

            return {
                draw_line_symbol(
                    ax,
                    line,
                    line_color=theme.line_color,
                    line_width=theme.thick_line_width,
                    ls="dashed",
                )
            }
        case PredicateType.EQUAL_ANGLES:
            return _draw_equal_angle_predicate(ax, predicate, symbols, theme=theme)
        case PredicateType.CONGRUENT:
            return cast(
                set[Artist],
                _draw_congruent_predicate(ax, predicate, theme=theme),
            )
        case PredicateType.PERPENDICULAR:
            return _draw_perpendicular_predicate(ax, predicate, symbols, theme=theme)
        case PredicateType.CYCLIC:
            circle = symbols.circles.circle_containing(set(predicate.points))
            if circle is None:
                raise ValueError(f"Circle not found for args: {predicate.points}")
            return {
                draw_circle_symbol(
                    ax,
                    circle,
                    line_color=theme.line_color,
                    line_width=theme.thick_line_width,
                )
            }
        case PredicateType.PARALLEL:
            return cast(
                set[Artist],
                _draw_parallel(ax, (predicate.line1, predicate.line2), theme=theme),
            )
        case PredicateType.SIMTRI_CLOCK | PredicateType.SIMTRI_REFLECT:
            triangles_sides = set(_draw_triangle(ax, *predicate.triangle1, theme=theme))
            return cast(set[Artist], triangles_sides)
        case PredicateType.EQUAL_RATIOS:
            return cast(
                set[Artist],
                _draw_eqratio_predicate(ax, predicate, symbols, theme=theme),
            )
        case _:
            return set()


def _draw_eqratio_predicate(
    ax: "Axes", predicate: EqRatio, symbols: SymbolsRegistry, theme: DrawTheme
) -> set[Line2D]:
    segments: set[Line2D] = set()
    segments.update(
        _draw_segment_num(
            ax,
            predicate.ratio1[0][0].num,
            predicate.ratio1[0][1].num,
            line_color=theme.line_color,
            line_width=theme.thick_line_width,
        )
    )
    segments.update(
        _draw_segment_num(
            ax,
            predicate.ratio1[0][0].num,
            predicate.ratio1[0][1].num,
            line_color=theme.perpendicular_color,
            line_width=theme.thick_line_width,
        )
    )
    segments.update(
        _draw_segment_num(
            ax,
            predicate.ratio2[0][0].num,
            predicate.ratio2[0][1].num,
            line_color=theme.line_color,
            line_width=theme.thick_line_width,
        )
    )
    segments.update(
        _draw_segment_num(
            ax,
            predicate.ratio2[0][0].num,
            predicate.ratio2[0][1].num,
            line_color=theme.perpendicular_color,
            line_width=theme.thick_line_width,
        )
    )
    return segments


def _draw_perpendicular_predicate(
    ax: "Axes", predicate: Perp, symbols: SymbolsRegistry, theme: DrawTheme
) -> set[Artist]:
    line1 = symbols.lines.line_containing(set(predicate.line1))
    if line1 is None:
        raise ValueError(f"Line not found for args: {predicate.line1}")
    line2 = symbols.lines.line_containing(set(predicate.line2))
    if line2 is None:
        raise ValueError(f"Line not found for args: {predicate.line2}")
    return _draw_perpendicular_lines(ax, line1, line2, theme=theme)


def _draw_perpendicular_lines(
    ax: "Axes", line1: LineSymbol, line2: LineSymbol, theme: DrawTheme
) -> set[Artist]:
    return {
        draw_perp_rectangle(
            ax,
            line1,
            line2,
            color=theme.perpendicular_color,
            zorder=0,
        ),
        draw_line_symbol(
            ax,
            line1,
            line_color=theme.line_color,
            line_width=theme.thin_line_width,
            zorder=0,
        ),
        draw_line_symbol(
            ax,
            line2,
            line_color=theme.line_color,
            line_width=theme.thin_line_width,
            zorder=0,
        ),
    }


def draw_perp_rectangle(
    ax: "Axes",
    line0: LineSymbol,
    line1: LineSymbol,
    color: str,
    **kwargs: Any,
) -> patches.Rectangle:
    args = fill_missing(
        {
            "color": color,
            "fill": False,
            "width": 0.025 * LENGTH_UNIT,
            "height": 0.025 * LENGTH_UNIT,
        },
        kwargs,
    )
    (o,) = line_line_intersection(line0.num, line1.num, ensure_point=True)
    ang0 = min(line0.num.angle(), line1.num.angle())
    rectangle = patches.Rectangle((o.x, o.y), angle=ang0 / np.pi * 180, **args)  # type: ignore
    ax.add_patch(rectangle)
    return rectangle


def _draw_congruent_predicate(
    ax: "Axes", predicate: Cong, theme: DrawTheme
) -> set[Line2D]:
    segments: set[Line2D] = set()
    for a, b in (predicate.segment1, predicate.segment2):
        segments.update(
            _draw_segment(
                ax,
                a,
                b,
                line_color=theme.line_color,
                line_width=theme.thick_line_width,
            )
        )
    return segments


def _draw_parallel(
    ax: "Axes",
    lines: tuple[Line, Line],
    theme: DrawTheme,
) -> set[Line2D]:
    setattr(ax, "para_color", (getattr(ax, "angle_color", 0) + 1) % len(PALETTE))
    seglen: int | float = 100
    lines_artists: set[Line2D] = set()

    for a, b in lines:
        lines_artists.update(
            _draw_segment(
                ax,
                a,
                b,
                line_color=theme.line_color,
                line_width=theme.thin_line_width,
                ls="dashed",
            )
        )
        seglen = min(seglen, a.num.distance(b.num))

    seglen /= 3.0
    for a, b in lines:
        d = b.num - a.num  # type: ignore
        d = d / abs(d)  # type: ignore
        d = d.rot90()
        if d.x < 0.0:
            d = -0.03 * d
        else:
            d = 0.03 * d
        p = b.num - a.num  # type: ignore
        p = p / abs(p)  # type: ignore
        p = p * (a.num.distance(b.num) - seglen) * 0.5
        lines_artists.update(
            _draw_segment_num(
                ax,
                a.num + d + p,
                b.num + d - p,
                line_color=PALETTE[ax.para_color],  # type: ignore
                line_width=theme.thin_line_width,
            )
        )
    return lines_artists


def _draw_equal_angle_predicate(
    ax: "Axes",
    predicate: EqAngle,
    symbols: SymbolsRegistry,
    theme: DrawTheme,
) -> set[Artist]:
    angles: list[tuple[LineSymbol, LineSymbol]] = []
    for angle in (predicate.angle1, predicate.angle2):
        l1 = symbols.lines.line_containing(set(angle[0]))
        l2 = symbols.lines.line_containing(set(angle[1]))
        assert l1 and l2
        angles.append((l1, l2))
    return _draw_equal_angles(ax, angles[0], angles[1], theme=theme)


def _draw_equal_angles(
    ax: "Axes",
    angle1: tuple[LineSymbol, LineSymbol],
    angle2: tuple[LineSymbol, LineSymbol],
    theme: DrawTheme,
) -> set[Artist]:
    setattr(ax, "angle_color", (getattr(ax, "angle_color", 0) + 1) % len(PALETTE))
    color = PALETTE[ax.angle_color]  # type: ignore
    r = (1 + 4 * random()) * LENGTH_UNIT / 50
    border_width = 0.05 * r
    artists: set[Artist] = set()
    for line1, line2 in (angle1, angle2):
        if not line1.num.is_parallel(line2.num):
            artists.add(
                _draw_angle_wedge(
                    ax,
                    line1,
                    line2,
                    color=color,
                    alpha=0.5,
                    r=r,
                    width=border_width,
                    zorder=1,
                )
            )
        artists.add(
            draw_line_symbol(
                ax,
                line1,
                line_color=theme.line_color,
                line_width=theme.thin_line_width,
                ls=":",
                zorder=1,
            )
        )
        artists.add(
            draw_line_symbol(
                ax,
                line2,
                line_color=theme.line_color,
                line_width=theme.thin_line_width,
                ls=":",
                zorder=1,
            )
        )
    return artists


def _draw_angle_wedge(
    ax: "Axes", line0: LineSymbol, line1: LineSymbol, **args: Any
) -> patches.Wedge:
    (o,) = line_line_intersection(line0.num, line1.num, ensure_point=True)
    ang0, ang1 = line0.num.angle(), line1.num.angle()
    if ang0 > ang1:
        ang0, ang1 = ang1, ang0
    if ang0 - ang1 + np.pi < ang1 - ang0:
        ang0, ang1 = ang1 - np.pi, ang0
    wedge = patches.Wedge(
        (o.x, o.y),
        theta1=ang0 / np.pi * 180,
        theta2=ang1 / np.pi * 180,
        **args,
    )
    ax.add_patch(wedge)
    return wedge


def draw_line_symbol(
    ax: "Axes", line: LineSymbol, line_color: str, line_width: float, **kwargs: Any
) -> AxLine:
    """Draw a line. Return the two extremities"""
    points: list[PointNum] = [p.num for p in line.points]
    p0, p1 = points[:2]
    return draw_line(ax, p0, p1, line_color=line_color, line_width=line_width, **kwargs)  # type: ignore


def draw_line(
    ax: "Axes",
    p0: PointNum,
    p1: PointNum,
    line_color: str,
    line_width: float,
    **kwargs: Any,
) -> AxLine:
    kwargs = fill_missing(kwargs, {"color": line_color, "lw": line_width})
    return ax.axline((p0.x, p0.y), (p1.x, p1.y), **kwargs)  # type: ignore


def _draw_triangle(
    ax: "Axes",
    a: Point,
    b: Point,
    c: Point,
    theme: DrawTheme,
) -> set[Line2D]:
    lines: set[Line2D] = set()
    lines.update(
        _draw_segment(
            ax,
            a,
            b,
            line_color=theme.line_color,
            line_width=theme.thin_line_width,
            ls="dashed",
        )
    )
    lines.update(
        _draw_segment(
            ax,
            b,
            c,
            line_color=theme.line_color,
            line_width=theme.thin_line_width,
            ls="dashed",
        )
    )
    lines.update(
        _draw_segment(
            ax,
            a,
            c,
            line_color=theme.line_color,
            line_width=theme.thin_line_width,
            ls="dashed",
        )
    )
    return lines


def _draw_segment(
    ax: "Axes",
    p0: Point,
    p1: Point,
    line_color: str,
    line_width: float,
    ls: str = "solid",
) -> list[Line2D]:
    lines = ax.plot(  # type: ignore
        (p0.num.x, p1.num.x),
        (p0.num.y, p1.num.y),
        color=line_color,
        lw=line_width,
        ls=ls,
    )
    return lines


def _draw_segment_num(
    ax: "Axes",
    p0: PointNum,
    p1: PointNum,
    line_color: str,
    line_width: float,
    **kwargs: Any,
) -> list[Line2D]:
    args = fill_missing(kwargs, {"color": line_color, "lw": line_width})
    lines = ax.plot((p0.x, p1.x), (p0.y, p1.y), **args)  # type: ignore
    return lines


def draw_free_perpendicular_symbol(
    ax: Axes,
    foot: PointNum,
    other_point: PointNum,
    theme: DrawTheme,
) -> Artist:
    vector = other_point - foot
    angle = vector.angle()
    rectangle = patches.Rectangle(
        (foot.x, foot.y),
        angle=angle / np.pi * 180,
        color=theme.perpendicular_color,
        fill=False,
        width=0.025,
        height=0.025,
    )  # type: ignore
    ax.add_artist(rectangle)
    return rectangle



from enum import Enum


class HeuristicName(str, Enum):
    LINE_INTERSECTIONS = "line_intersections"
    CENTERS = "centers"
    ANGLE_VERTICES = "angle_vertices"
    MIDPOINT = "midpoint"
    EQDISTANCE = "eqdistance"
    FOOT = "foot"
    REFLECT_ON_CENTER = "reflect_on_center"
    THREE_CIRCUMCIRCLES_FOR_EQANGLE_GOAL = "three_circumcircles_for_eqangle_goal"



"""Deductors are alternative ways to get deductions from other sources than rules and symbols merges.

The main example of deductor is the algebraic reasoning, which is used to get deductions from angle chasing and ratio chasing.
"""

from enum import Enum


class ARReason(str, Enum):
    ANGLE_CHASING = "Angle Chasing"
    RATIO_CHASING = "Ratio Chasing"



"""Definitions of points on lines, circles, etc."""

from newclid.jgex.clause import JGEXClause
from newclid.jgex.constructions._index import JGEXConstructionName
from newclid.jgex.definition import JGEXDefinition, SketchConstruction, v
from newclid.jgex.sketch import (
    sketch_aline,
    sketch_aline0,
    sketch_bline,
    sketch_circle,
    sketch_cyclic,
    sketch_dia,
    sketch_function_name,
    sketch_line,
    sketch_pline,
    sketch_tline,
)

# Point on Line

ON_LINE = JGEXDefinition(
    name=JGEXConstructionName.ON_LINE,
    args=(v("x"), v("a"), v("b")),
    rely_on_points={v("x"): (v("a"), v("b"))},
    requirements=JGEXClause.from_str("diff a b")[0],
    clauses=JGEXClause.from_str("coll x a b"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_line.__name__), args=(v("a"), v("b"))
        ),
    ),
    input_points=(v("a"), v("b")),
    output_points=(v("x"),),
)

ON_PARA_LINE = JGEXDefinition(
    name=JGEXConstructionName.ON_PARA_LINE,
    args=(v("x"), v("a"), v("b"), v("c")),
    rely_on_points={v("x"): (v("a"), v("b"), v("c"))},
    requirements=JGEXClause.from_str("diff b c, ncoll a b c")[0],
    clauses=JGEXClause.from_str("para x a b c"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_pline.__name__),
            args=(v("a"), v("b"), v("c")),
        ),
    ),
    input_points=(v("a"), v("b"), v("c")),
    output_points=(v("x"),),
)

ON_PARA_LINE0 = JGEXDefinition(
    name=JGEXConstructionName.ON_PARA_LINE0,
    args=(v("x"), v("a"), v("b"), v("c")),
    rely_on_points={v("x"): (v("a"), v("b"), v("c"))},
    requirements=JGEXClause.from_str("diff b c")[0],
    clauses=JGEXClause.from_str("para x a b c"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_pline.__name__),
            args=(v("a"), v("b"), v("c")),
        ),
    ),
    input_points=(v("a"), v("b"), v("c")),
    output_points=(v("x"),),
)

ON_PERP_LINE = JGEXDefinition(
    name=JGEXConstructionName.ON_PERP_LINE,
    args=(v("x"), v("a"), v("b"), v("c")),
    rely_on_points={v("x"): (v("a"), v("b"), v("c"))},
    requirements=JGEXClause.from_str("diff b c")[0],
    clauses=JGEXClause.from_str("perp x a b c"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_tline.__name__),
            args=(v("a"), v("b"), v("c")),
        ),
    ),
    input_points=(v("a"), v("b"), v("c")),
    output_points=(v("x"),),
)


ON_ALINE = JGEXDefinition(
    name=JGEXConstructionName.ON_ALINE,
    args=(v("x"), v("a"), v("b"), v("c"), v("d"), v("e")),
    rely_on_points={v("x"): (v("a"), v("b"), v("c"), v("d"), v("e"))},
    requirements=JGEXClause.from_str("ncoll c d e")[0],
    clauses=JGEXClause.from_str("eqangle a x a b d c d e"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_aline.__name__),
            args=(v("e"), v("d"), v("c"), v("b"), v("a")),
        ),
    ),
    input_points=(v("a"), v("b"), v("c"), v("d"), v("e")),
    output_points=(v("x"),),
)


ON_ALINE0 = JGEXDefinition(
    name=JGEXConstructionName.ON_ALINE0,
    args=(v("x"), v("a"), v("b"), v("c"), v("d"), v("e"), v("f"), v("g")),
    rely_on_points={v("x"): (v("a"), v("b"), v("c"), v("d"), v("e"), v("f"), v("g"))},
    requirements=JGEXClause.from_str("ncoll a b c d")[0],
    clauses=JGEXClause.from_str("eqangle a b c d e f g x"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_aline0.__name__),
            args=(v("a"), v("b"), v("c"), v("d"), v("e"), v("f"), v("g")),
        ),
    ),
    input_points=(v("a"), v("b"), v("c"), v("d"), v("e"), v("f"), v("g")),
    output_points=(v("x"),),
)


ON_BLINE = JGEXDefinition(
    name=JGEXConstructionName.ON_BLINE,
    args=(v("x"), v("a"), v("b")),
    rely_on_points={v("x"): (v("a"), v("b"))},
    requirements=JGEXClause.from_str("diff a b")[0],
    clauses=JGEXClause.from_str("cong x a x b, eqangle a x a b b a b x"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_bline.__name__), args=(v("a"), v("b"))
        ),
    ),
    input_points=(v("a"), v("b")),
    output_points=(v("x"),),
)


ON_CIRCLE = JGEXDefinition(
    name=JGEXConstructionName.ON_CIRCLE,
    args=(v("x"), v("o"), v("a")),
    rely_on_points={v("x"): (v("o"), v("a"))},
    requirements=JGEXClause.from_str("diff o a")[0],
    clauses=JGEXClause.from_str("cong o x o a"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_circle.__name__),
            args=(v("o"), v("o"), v("a")),
        ),
    ),
    input_points=(v("o"), v("a")),
    output_points=(v("x"),),
)

ON_CIRCUM = JGEXDefinition(
    name=JGEXConstructionName.ON_CIRCUM,
    args=(v("x"), v("a"), v("b"), v("c")),
    rely_on_points={v("x"): (v("a"), v("b"), v("c"))},
    requirements=JGEXClause.from_str("ncoll a b c")[0],
    clauses=JGEXClause.from_str("cyclic a b c x"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_cyclic.__name__),
            args=(v("a"), v("b"), v("c")),
        ),
    ),
    input_points=(v("a"), v("b"), v("c")),
    output_points=(v("x"),),
)

ON_DIA = JGEXDefinition(
    name=JGEXConstructionName.ON_DIA,
    args=(v("x"), v("a"), v("b")),
    rely_on_points={v("x"): (v("a"), v("b"))},
    requirements=JGEXClause.from_str("diff a b")[0],
    clauses=JGEXClause.from_str("perp x a x b"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_dia.__name__), args=(v("a"), v("b"))
        ),
    ),
    input_points=(v("a"), v("b")),
    output_points=(v("x"),),
)

POINT_ON_OBJECT_CONSTRUCTIONS = [
    ON_LINE,
    ON_PARA_LINE,
    ON_PARA_LINE0,
    ON_PERP_LINE,
    ON_ALINE,
    ON_ALINE0,
    ON_BLINE,
    ON_CIRCLE,
    ON_CIRCUM,
    ON_DIA,
]



from sympy.core.backend import Symbol as SympySymbol  # type: ignore

from newclid.deductors.sympy_ar.ar_table import ARTable
from newclid.numerical import close_enough
from newclid.symbols.points_registry import Segment


class RatiosTable:
    def __init__(self, ar_table: ARTable) -> None:
        self.inner_table = ar_table

        self.expected_lconsts: dict[SympySymbol, float] = {}
        self.expected_rconsts: dict[tuple[SympySymbol, SympySymbol], float] = {}
        self.expected_congs: set[tuple[SympySymbol, SympySymbol]] = set()
        self.expected_eqratios: set[
            tuple[SympySymbol, SympySymbol, SympySymbol, SympySymbol]
        ] = set()

        self.segment_to_sympy_symbol: dict[Segment, SympySymbol] = {}
        self.sympy_symbol_to_str_symbol: dict[SympySymbol, Segment] = {}

    def segment_log_length(self, segment: Segment) -> SympySymbol:
        p0, p1 = segment
        if p0.name > p1.name:
            p0, p1 = p1, p0

        if (p0, p1) not in self.segment_to_sympy_symbol:
            name = f"log_seg_length_{p0.name}_{p1.name}"
            length_symbol = SympySymbol(name)
            self.segment_to_sympy_symbol[(p0, p1)] = length_symbol
            self.sympy_symbol_to_str_symbol[length_symbol] = (p0, p1)
            lenght_value = p0.num.distance(p1.num)
            self._add_log_length(length_symbol, lenght_value)

        return self.segment_to_sympy_symbol[(p0, p1)]

    def _add_log_length(self, length_symbol: SympySymbol, numerical_value: float):
        for other_symbol, other_length in self.expected_lconsts.items():
            new_ratio = (length_symbol, other_symbol)
            new_ratio_value = numerical_value / other_length

            if close_enough(numerical_value, other_length):
                self.expected_congs.add((length_symbol, other_symbol))

            for other_ratio, other_ratio_value in self.expected_rconsts.items():
                if close_enough(new_ratio_value, other_ratio_value):
                    self.expected_eqratios.add((*new_ratio, *other_ratio))

            self.expected_rconsts[new_ratio] = new_ratio_value

            self.expected_rconsts[(other_symbol, length_symbol)] = 1 / new_ratio_value

        self.expected_lconsts[length_symbol] = numerical_value



from __future__ import annotations

from typing import TYPE_CHECKING, NewType

from pydantic import BaseModel

from newclid.jgex.clause import JGEXClause, JGEXConstruction
from newclid.problem import PredicateConstruction

if TYPE_CHECKING:
    from newclid.jgex.to_newclid import JGEXClauseConsequences

JGEXSetupId = NewType("JGEXSetupId", str)


class PredicateConstructionInSetup(BaseModel):
    id: JGEXSetupId
    construction: PredicateConstruction

    def __str__(self) -> str:
        return f"[{self.id}] {str(self.construction)}"


class JGEXClauseInProof(BaseModel):
    """Constructions of a clause and their resulting predicates."""

    added_points: tuple[str, ...]
    constructions: tuple[JGEXConstruction, ...]
    predicates: tuple[PredicateConstructionInSetup, ...]

    def __str__(self) -> str:
        points = " ".join(self.added_points)
        constructions = ", ".join(str(c) for c in self.constructions)
        predicates = ", ".join(str(s_in_proof) for s_in_proof in self.predicates)
        return f"{points} = {constructions} ({predicates})"


class JGEXSetupData(BaseModel):
    setup_clauses: list[JGEXClauseInProof]
    """The necessary problem constuctions and their consequence predicates."""
    aux_clauses: list[JGEXClauseInProof]
    """The auxiliary constructions and their consequence predicates."""
    goals: list[PredicateConstructionInSetup]
    """Goal predicates."""


def jgex_clauses_to_setup_data(
    setup_clauses: list[JGEXClause],
    aux_clauses: list[JGEXClause],
    goals: list[PredicateConstruction],
    clauses_consequences: dict[JGEXClause, JGEXClauseConsequences],
) -> tuple[JGEXSetupData, dict[JGEXSetupId, PredicateConstruction]]:
    predicates_ids: dict[JGEXSetupId, PredicateConstruction] = {}
    construction_predicates: dict[JGEXSetupId, PredicateConstruction] = {}
    setup_clauses_in_proof = _assign_clauses_predicates(
        clauses=setup_clauses,
        clauses_consequences=clauses_consequences,
        predicates_ids=predicates_ids,
        construction_predicates=construction_predicates,
    )
    aux_clauses_in_proof = _assign_clauses_predicates(
        clauses=aux_clauses,
        clauses_consequences=clauses_consequences,
        predicates_ids=predicates_ids,
        construction_predicates=construction_predicates,
    )
    goals_in_proof = _assign_goals_predicates(goals, predicates_ids)
    setup_data = JGEXSetupData(
        setup_clauses=setup_clauses_in_proof,
        aux_clauses=aux_clauses_in_proof,
        goals=goals_in_proof,
    )
    return setup_data, predicates_ids


def _assign_goals_predicates(
    goals: list[PredicateConstruction],
    predicates_ids: dict[JGEXSetupId, PredicateConstruction],
) -> list[PredicateConstructionInSetup]:
    goals_in_proof: list[PredicateConstructionInSetup] = []
    for index, goal in enumerate(goals):
        goal_tag = JGEXSetupId(f"G{index}")
        if goal_tag in predicates_ids:
            raise ValueError(f"Goal {goal} already has a predicate id")
        predicates_ids[goal_tag] = goal
        goals_in_proof.append(
            PredicateConstructionInSetup(id=goal_tag, construction=goal)
        )
    return goals_in_proof


def _assign_clauses_predicates(
    clauses: list[JGEXClause],
    clauses_consequences: dict[JGEXClause, JGEXClauseConsequences],
    predicates_ids: dict[JGEXSetupId, PredicateConstruction],
    construction_predicates: dict[JGEXSetupId, PredicateConstruction],
) -> list[JGEXClauseInProof]:
    return [
        _assign_clause_predicates(
            clause, clauses_consequences, predicates_ids, construction_predicates
        )
        for clause in clauses
    ]


def _assign_clause_predicates(
    clause: JGEXClause,
    clauses_consequences: dict[JGEXClause, JGEXClauseConsequences],
    predicates_ids: dict[JGEXSetupId, PredicateConstruction],
    construction_predicates: dict[JGEXSetupId, PredicateConstruction],
) -> JGEXClauseInProof:
    clause_predicates_in_proof: list[PredicateConstructionInSetup] = []
    constructions = clauses_consequences[clause].construction_consequences
    for construction_consequence in constructions:
        for predicate_construction in construction_consequence.added_predicates:
            n_predicates = len(construction_predicates)
            tag = JGEXSetupId(f"C{n_predicates}")
            if tag in predicates_ids:
                raise ValueError(
                    f"Tag {tag} already has predicate {predicates_ids[tag]} attributed to it"
                )
            predicates_ids[tag] = predicate_construction
            construction_predicates[tag] = predicate_construction
            clause_predicates_in_proof.append(
                PredicateConstructionInSetup(
                    id=tag, construction=predicate_construction
                )
            )
    return JGEXClauseInProof(
        constructions=clause.constructions,
        predicates=tuple(clause_predicates_in_proof),
        added_points=clause.points,
    )



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from pydantic import BaseModel

if TYPE_CHECKING:
    from newclid.predicate_types import PredicateArgument
    from newclid.proof_state import ProofState
    from newclid.symbols.symbols_registry import LineOrCircle, SymbolsRegistry


class PredicateInterface(ABC, BaseModel):
    """
    When the args are passed in functions other than parse and to_tokens,
    the orders are guaranteed to be canonique.
    """

    @staticmethod
    @abstractmethod
    def preparse(
        args: tuple[PredicateArgument, ...],
    ) -> tuple[PredicateArgument, ...] | None:
        """Preparse the predicate arguments."""

    @abstractmethod
    def check_numerical(self) -> bool:
        """Check numerically the predicate."""

    @abstractmethod
    def to_tokens(self) -> tuple[PredicateArgument, ...]:
        """Convert the predicate to a tuple of strings."""

    def add(self, proof_state: ProofState) -> tuple[PredicateInterface, ...]:
        """Add the predicate to the proof state.

        Return a tuple of predicates that are direct consequences of the predicate by definition.
        """
        return ()

    def check(self, proof_state: ProofState) -> bool | None:
        """Check symbolically the predicate in the current proof state.

        If the predicate cannot be decided, return None.
        """
        return None

    def symbols(self, symbols: SymbolsRegistry) -> tuple[LineOrCircle, ...]:
        """Make symbols for the predicate in the symbols graph."""
        return ()

    def __hash__(self) -> int:
        return hash(str(self))

    def __repr__(self) -> str:
        return f"{self.predicate_type.value} {' '.join(self.to_tokens())}"  # type: ignore


class NumericalPredicate(PredicateInterface):
    """Predicate that can only be checked numerically."""

    def check(self, proof_state: ProofState) -> bool:
        """Numerical predicates are always true symbolically."""
        return True



import itertools
import logging
from typing import Literal

import numpy
from pydantic import BaseModel

from newclid.heuristics._index import HeuristicName
from newclid.heuristics._interface import Heuristic, HeuristicSetup
from newclid.heuristics.alphabet import get_available_from_alphabet
from newclid.jgex.clause import JGEXClause
from newclid.predicate_types import PredicateArgument

LOGGER = logging.getLogger(__name__)


class LineIntersectionsHeuristicConfig(BaseModel):
    heuristic_name: Literal[HeuristicName.LINE_INTERSECTIONS] = (
        HeuristicName.LINE_INTERSECTIONS
    )


class LineIntersectionsHeuristic(Heuristic):
    """Add intersection points between lines."""

    def new_clauses(
        self,
        setup: HeuristicSetup,
        max_new_points: int,
        rng: numpy.random.Generator,
    ) -> list[JGEXClause]:
        pairs_of_lines = list(itertools.combinations(setup.lines, 2))
        added_points: list[PredicateArgument] = []
        new_clauses: list[JGEXClause] = []

        for _ in range(max_new_points):
            if not pairs_of_lines:
                break

            pair_index = rng.choice(len(pairs_of_lines))
            line1, line2 = pairs_of_lines.pop(pair_index)

            point_1 = line1.points[0]
            point_2 = line1.points[1]

            point_3 = line2.points[0]
            point_4 = line2.points[1]

            new_point = get_available_from_alphabet(list(setup.points) + added_points)
            if new_point is None:
                LOGGER.warning(
                    "No available points left in the alphabet, skipping intersection addition."
                )
                break

            added_points.append(new_point)

            new_clause_string = f"{new_point} = on_line {new_point} {point_1} {point_2}, on_line {new_point} {point_3} {point_4}"
            new_clause = JGEXClause.from_str(new_clause_string)[0]
            new_clauses.append(new_clause)

        return new_clauses



from matplotlib.artist import Artist
from matplotlib.axes import Axes

from newclid.deductors import ARReason
from newclid.draw.predicates import draw_predicate
from newclid.draw.theme import DrawTheme
from newclid.justifications.justification import ARDeduction
from newclid.symbols.symbols_registry import SymbolsRegistry


def draw_ar_application(
    ax: Axes,
    ar_deduction: ARDeduction,
    symbols: SymbolsRegistry,
    theme: DrawTheme,
) -> set[Artist]:
    match ar_deduction.ar_reason:
        case ARReason.ANGLE_CHASING:
            return draw_deps_predicates(ax, ar_deduction, symbols, theme=theme)
        case ARReason.RATIO_CHASING:
            return draw_deps_predicates(ax, ar_deduction, symbols, theme=theme)
        case _:  # pyright: ignore
            raise NotImplementedError(
                f"Cannot draw AR application: {ar_deduction.ar_reason}"
            )


def draw_deps_predicates(
    ax: Axes,
    ar_deduction: ARDeduction,
    symbols: SymbolsRegistry,
    theme: DrawTheme,
) -> set[Artist]:
    if ar_deduction.ar_premises is None:
        raise ValueError(
            f"No premises to apply rule {ar_deduction.ar_reason} in dependency: {ar_deduction}"
        )

    artists: set[Artist] = set()
    artists.update(draw_predicate(ax, ar_deduction.predicate, symbols, theme=theme))
    for premise in ar_deduction.ar_premises:
        artists.update(draw_predicate(ax, premise.predicate, symbols, theme=theme))
    return artists



"""JGEX problems from the International Mathematical Olympiad."""

from pydantic_core import Url

from newclid.jgex.formulation import JGEXFormulation
from newclid.problems.problem import (
    Competition,
    CompetitionIds,
    Problem,
    ProblemFormulation,
)

IMO = Competition(
    id=CompetitionIds.IMO,
    name="International Mathematical Olympiad",
    url=Url("https://www.imo-official.org/"),
)

IMO_2000_P1 = Problem(
    source=IMO,
    name="2000_p1",
    natural_language_formulation="""AB is tangent to the circles CAMN and NMBD.
M lies between C and D on the line CD, and CD is parallel to AB.
The chords NA and CM meet at P""; the chords NB and MD meet at Q.
The rays CA and DB meet at E. Prove that P E = QE.""",
    formulations=[
        ProblemFormulation(
            name="ag",
            formulation=JGEXFormulation.from_text(
                "a b = segment a b"
                "; g1 = on_tline g1 a a b"
                "; g2 = on_tline g2 b b a"
                "; m = on_circle m g1 a, on_circle m g2 b"
                "; n = on_circle n g1 a, on_circle n g2 b"
                "; c = on_pline c m a b, on_circle c g1 a"
                "; d = on_pline d m a b, on_circle d g2 b"
                "; p = on_line p a n, on_line p c d"
                "; q = on_line q b n, on_line q c d"
                "; e = on_line e a c, on_line e b d"
                " ? cong e p e q"
            ),
        )
    ],
)

IMO_2025_P2 = Problem(
    source=IMO,
    name="2025_p2",
    natural_language_formulation="""Let Ω and Γ be circles with centres Mand N, respectively, such that the radius of Ω is less than the radius of Γ.
Suppose circles Ω and Γ intersect at two distinct points A and B.
Line MN intersects Ω at C and Γ at D, such that points C, M, N and D lie on the line in that order.
Let P be the circumcentre of triangle ACD. Line AP intersects Ω again at E ≠ A.
Line AP intersects Γ again at F ≠ A.
Let H be the orthocentre of triangle PMN.
Prove that the line through H parallel to AP is tangent to the circumcircle of triangle BEF.""",
    formulations=[
        ProblemFormulation(
            name="2025_p2",
            formulation=JGEXFormulation.from_text(
                "a m n = triangle a m n"
                "; b = on_circle b m a, on_circle b n a"
                "; c = on_line c m n, on_circle c m a"
                "; d = on_line d m n, on_circle d n a"
                "; p = circle p a c d"
                "; e = on_line e a p, on_circle e m a"
                "; f = on_line f a p, on_circle f n a"
                "; h = on_tline h p m n, on_tline h m p n"
                "; o = circle o b e f"
                "; x = on_dia x o h, on_pline x h a p"
                " | w = circle w d e f"
                " ? cong o x o b"
                "; obtuse_angle c m n"
                "; obtuse_angle m n d"
            ),
        ),
    ],
)

ALL_IMO_PROBLEMS = [
    IMO_2000_P1,
    IMO_2025_P2,
]



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from matplotlib.axes import Axes
from matplotlib.collections import PathCollection
from matplotlib.figure import Figure
from matplotlib.text import Annotation

from newclid.draw.jgex_constructions import draw_jgex_problem_clauses
from newclid.draw.predicates import draw_predicate
from newclid.draw.theme import DrawTheme
from newclid.proof_justifications import goals_justifications
from newclid.symbols.points_registry import Point

if TYPE_CHECKING:
    from newclid.jgex.formulation import JGEXFormulation
    from newclid.proof_state import ProofState


def init_figure(ax: Axes, points: list[Point]) -> None:
    ax.set_facecolor((0.0, 0.0, 0.0))
    ax.set_aspect("equal")

    xmin, xmax = min(p.num.x for p in points), max(p.num.x for p in points)
    x_range = xmax - xmin
    ymin, ymax = min(p.num.y for p in points), max(p.num.y for p in points)
    y_range = ymax - ymin
    side_range = max(x_range, y_range)

    ax.set_xlim(xmin - 0.1 * side_range, xmin + 1.1 * side_range)
    ax.set_ylim(ymin - 0.1 * side_range, ymin + 1.1 * side_range)
    ax.get_xaxis().set_visible(False)
    ax.get_yaxis().set_visible(False)


def draw_figure(
    fig: Figure,
    ax: Axes,
    proof: "ProofState",
    jgex_problem: JGEXFormulation | None,
    theme: DrawTheme,
    save_to: Path | None = None,
    format: str = "svg",
) -> tuple[Figure, Axes]:
    """Draw everything on the same canvas."""
    symbols_registry = proof.symbols
    points: list[Point] = list(symbols_registry.points.name_to_point.values())
    init_figure(ax, points)
    ax.set_aspect("equal", adjustable="datalim")

    if jgex_problem is not None:
        draw_jgex_problem_clauses(
            ax,
            jgex_problem=jgex_problem,
            symbols_registry=symbols_registry,
            theme=theme,
        )

    if proof.check_goals():
        justifications, _ = goals_justifications(proof.goals, proof_state=proof)
        predicates_to_draw = [
            justification.predicate for justification in justifications
        ]
    else:
        predicates_to_draw = proof.graph.predicates

    for predicate in predicates_to_draw:
        draw_predicate(ax, predicate, symbols_registry, theme=theme)
    for p in points:
        draw_point(ax, p, point_color=theme.point_color, zorder=100)

    if save_to is not None:
        save_to.parent.mkdir(parents=True, exist_ok=True)
        fig.savefig(save_to, format=format)  # type: ignore

    return fig, ax


def draw_point(
    ax: "Axes",
    p: Point,
    point_color: str,
    fontsize: float = 20,
    point_size: float = 25.0,
    zorder: float = 100.0,
) -> tuple[PathCollection, Annotation]:
    """draw a point."""
    point_artist = ax.scatter(  # type: ignore
        p.num.x,
        p.num.y,
        color=point_color,
        s=point_size,
        zorder=zorder,
        rasterized=True,
    )
    annotation_artist = ax.annotate(  # type: ignore
        str(p),
        (p.num.x, p.num.y),
        color=point_color,
        fontsize=fontsize,
        zorder=zorder,
        rasterized=True,
    )
    return point_artist, annotation_artist



from __future__ import annotations

from newclid.numerical import ATOM
from newclid.numerical.geometries import PointNum


def same_clock(
    a: PointNum, b: PointNum, c: PointNum, d: PointNum, e: PointNum, f: PointNum
) -> bool:
    """
    a, b, c; d, e, f are of the same clock and they are not colinear
    """
    clock_left = clock(a, b, c)
    clock_right = clock(d, e, f)
    return (clock_left > ATOM and clock_right > ATOM) or (
        clock_left < -ATOM and clock_right < -ATOM
    )


def clock(a: PointNum, b: PointNum, c: PointNum):
    ab = b - a
    ac = c - a
    return ab.x * ac.y - ab.y * ac.x



from __future__ import annotations

from typing import Literal

from newclid.numerical import ATOM
from newclid.predicate_types import PredicateArgument
from newclid.predicates._index import PredicateType
from newclid.predicates._interface import NumericalPredicate
from newclid.symbols.points_registry import Point


class ObtuseAngle(NumericalPredicate):
    """obtuse_angle a b c

    Represents that the angle `a b c` is an obtuse angle.
    If `a`, `b`, `c` are collinear, this predicate represents that `b` is between `a` and `c`.

    Numerical only.
    """

    predicate_type: Literal[PredicateType.OBTUSE_ANGLE] = PredicateType.OBTUSE_ANGLE

    head1: Point
    corner: Point
    head2: Point

    @staticmethod
    def preparse(
        args: tuple[PredicateArgument, ...],
    ) -> tuple[PredicateArgument, ...] | None:
        a, b, c = args
        if a == b or b == c or a == c:
            return None
        (a, c) = sorted((a, c))
        return (a, b, c)

    def check_numerical(self) -> bool:
        ray_1 = self.head1.num - self.corner.num
        ray_2 = self.head2.num - self.corner.num
        return ray_1.dot(ray_2) < -ATOM

    def to_tokens(self) -> tuple[PredicateArgument, ...]:
        return tuple(
            PredicateArgument(p.name) for p in (self.head1, self.corner, self.head2)
        )

    def __str__(self) -> str:
        return f"∠({self.head2}{self.corner}{self.head1}) is an obtuse angle"



from __future__ import annotations

from typing import Literal

import numpy as np

from newclid.numerical import close_enough
from newclid.predicate_types import PredicateArgument
from newclid.predicates._index import PredicateType
from newclid.predicates._interface import PredicateInterface
from newclid.symbols.points_registry import Segment
from newclid.symbols.symbols_registry import LineOrCircle, SymbolsRegistry
from newclid.tools import reshape

AngleArgs = tuple[
    PredicateArgument, PredicateArgument, PredicateArgument, PredicateArgument
]


class EqAngle(PredicateInterface):
    """eqangle AB CD EF GH -
    Represent that one can rigidly move the crossing of lines AB and CD
    to get on top of the crossing of EF and GH, respectively (no reflections allowed).

    In particular, eqangle AB CD CD AB is only true if AB is perpendicular or parallel to CD.
    """

    predicate_type: Literal[PredicateType.EQUAL_ANGLES] = PredicateType.EQUAL_ANGLES
    angle1: tuple[Segment, Segment]
    angle2: tuple[Segment, Segment]

    @staticmethod
    def preparse(
        args: tuple[PredicateArgument, ...],
    ) -> tuple[PredicateArgument, ...] | None:
        groups: list[AngleArgs] = []
        groups1: list[AngleArgs] = []
        if len(args) % 4:
            return None
        for a, b, c, d in reshape(args, 4):
            if a == b or c == d:
                return None
            a, b = sorted((a, b))
            c, d = sorted((c, d))
            groups.append((a, b, c, d))
            groups1.append((c, d, a, b))
        return sum(min(sorted(groups), sorted(groups1)), ())

    def check_numerical(self) -> bool:
        (a, b), (c, d) = self.angle1
        (e, f), (g, h) = self.angle2

        angle1 = ((d.num - c.num).angle() - (b.num - a.num).angle()) % np.pi
        angle2 = ((h.num - g.num).angle() - (f.num - e.num).angle()) % np.pi

        if (
            not close_enough(angle1, angle2)
            and not close_enough(angle1, angle2 + np.pi)
            and not close_enough(angle1, angle2 - np.pi)
        ):
            return False
        return True

    def symbols(self, symbols: SymbolsRegistry) -> tuple[LineOrCircle, ...]:
        (a, b), (c, d) = self.angle1
        l1 = symbols.lines.line_thru_pair(a, b)
        l2 = symbols.lines.line_thru_pair(c, d)
        (e, f), (g, h) = self.angle2
        l3 = symbols.lines.line_thru_pair(e, f)
        l4 = symbols.lines.line_thru_pair(g, h)
        return (l1, l2, l3, l4)

    @staticmethod
    def to_constructive(point: str, args: tuple[str, ...]) -> str:
        a, b, c, d, e, f = args

        if point in [d, e, f]:
            a, b, c, d, e, f = d, e, f, a, b, c

        x, b, y, c, d = b, c, e, d, f
        if point == b:
            a, b, c, d = b, a, d, c

        if point == d and x == y:  # x p x b = x c x p
            return f"angle_bisector {point} {b} {x} {c}"

        if point == x:
            return f"eqangle3 {x} {a} {b} {y} {c} {d}"

        return f"on_aline {a} {x} {b} {c} {y} {d}"

    def to_tokens(self) -> tuple[PredicateArgument, ...]:
        return tuple(
            PredicateArgument(p.name)
            for ((a, b), (c, d)) in [self.angle1, self.angle2]
            for p in (a, b, c, d)
        )

    def __str__(self) -> str:
        return " = ".join(
            f"∠({a}{b},{c}{d})" for ((a, b), (c, d)) in [self.angle1, self.angle2]
        )



from newclid.ggb.read_commands import (
    GGBAngularBisectorLinesCommand,
    GGBAngularBisectorPointsCommand,
    GGBCenterCommand,
    GGBCircleCommand,
    GGBCircumcircleCommand,
    GGBCommand,
    GGBCompassCommand,
    GGBIntersectCommand,
    GGBLineCommand,
    GGBOrthogonalLineCommand,
    GGBParaCommand,
    GGBReadCommandType,
    GGBTangentOnCircleCommand,
    GGBTangentOutsideCircleCommand,
)
from newclid.ggb.read_elements import GGBConic, GGBLine


def apply_commands_effects_on_elements(
    commands: list[GGBCommand],
    lines: dict[str, GGBLine],
    conics: dict[str, GGBConic],
) -> None:
    for command in commands:
        _apply_command_effect(command, lines, conics)


def _apply_command_effect(
    command: GGBCommand,
    lines: dict[str, GGBLine],
    conics: dict[str, GGBConic],
) -> None:
    match command.construction_type:
        case GGBReadCommandType.Line:
            _add_points_on_line_from_line_command(command, lines)
        case GGBReadCommandType.Intersect:
            _add_points_on_objects_in_intersect_command(command, lines, conics)
        case GGBReadCommandType.OrthogonalLine:
            _add_points_and_line_on_line_from_orthogonal_line_command(command, lines)
        case GGBReadCommandType.ParallelLine:
            _add_points_and_line_on_line_from_parallel_line_command(command, lines)
        case GGBReadCommandType.Circle:
            _add_point_and_center_on_conic_from_circle_command(command, conics)
            _add_radius_from_circle_command(command, conics)
        case GGBReadCommandType.Circumcircle:
            _add_points_on_conic_from_circumcircle_command(command, conics)
        case GGBReadCommandType.Eqdistance:
            _add_point_and_center_from_compass_command(command, conics)
            _add_radius_from_compass_command(command, conics)
        case GGBReadCommandType.AddPointOn:
            _add_point_on_object_command(
                command.point, command.object_name, lines, conics
            )
        case GGBReadCommandType.Center:
            _add_center_to_circle_from_center_command(command, conics)
        case GGBReadCommandType.AngularBisector:
            _add_points_bisecting_lines_from_bisector_command(command, lines)
        case GGBReadCommandType.AngularBisectorPoints:
            _add_bisecting_angles_from_bisector_command(command, lines)
        case GGBReadCommandType.TangentOnCircle:
            _add_tangency_elements_from_tangent_on_circle_command(command, lines)
        case GGBReadCommandType.TangentOutsideCircle:
            _add_tangency_elements_from_tangent_outside_circle_command(command, lines)
        case GGBReadCommandType.AngularBisectorPoints:
            _add_vertex_to_bisector_from_angular_bisector_command(command, lines)
        case _:
            pass


def _add_points_on_line_from_line_command(
    command: GGBLineCommand, lines: dict[str, GGBLine]
) -> None:
    lines[command.line].points.extend([command.point_start, command.point_end])


def _add_points_on_conic_from_circumcircle_command(
    command: GGBCircumcircleCommand, conics: dict[str, GGBConic]
) -> None:
    conics[command.circle].points.extend(
        [command.point_a, command.point_b, command.point_c]
    )


def _add_point_and_center_from_compass_command(
    command: GGBCompassCommand, conics: dict[str, GGBConic]
) -> None:
    conics[command.circle].center = command.center


def _add_radius_from_compass_command(
    command: GGBCompassCommand, conics: dict[str, GGBConic]
) -> None:
    conics[command.circle].radius = command.segment_radius


def _add_points_and_line_on_line_from_orthogonal_line_command(
    command: GGBOrthogonalLineCommand, lines: dict[str, GGBLine]
) -> None:
    lines[command.perpendicular_line].points.append(command.point)


def _add_points_and_line_on_line_from_parallel_line_command(
    command: GGBParaCommand, lines: dict[str, GGBLine]
) -> None:
    lines[command.parallel_line].points.append(command.point_start)


def _add_points_on_objects_in_intersect_command(
    command: GGBIntersectCommand, lines: dict[str, GGBLine], conics: dict[str, GGBConic]
) -> None:
    _add_point_on_object_command(command.point, command.object_1, lines, conics)
    _add_point_on_object_command(command.point, command.object_2, lines, conics)


def _add_point_and_center_on_conic_from_circle_command(
    command: GGBCircleCommand, conics: dict[str, GGBConic]
) -> None:
    conics[command.circle].points.append(command.point_b)
    conics[command.circle].center = command.center


def _add_radius_from_circle_command(
    command: GGBCircleCommand, conics: dict[str, GGBConic]
) -> None:
    conics[command.circle].radius = (command.center, command.point_b)


def _add_point_on_object_command(
    point: str,
    object_name: str,
    lines: dict[str, GGBLine],
    conics: dict[str, GGBConic],
) -> None:
    if object_name in lines:
        lines[object_name].points.append(point)
    elif object_name in conics:
        conics[object_name].points.append(point)


def _add_center_to_circle_from_center_command(
    command: GGBCenterCommand, conics: dict[str, GGBConic]
) -> None:
    conics[command.circle].center = command.point


def _add_points_bisecting_lines_from_bisector_command(
    command: GGBAngularBisectorLinesCommand, lines: dict[str, GGBLine]
) -> None:
    for point in lines[command.line_1].points:
        if point in lines[command.line_2].points:
            lines[command.line_bisector].points.append(point)


def _add_bisecting_angles_from_bisector_command(
    command: GGBAngularBisectorPointsCommand, lines: dict[str, GGBLine]
) -> None:
    lines[command.line_bisector].points.extend(command.corner)


def _add_tangency_elements_from_tangent_on_circle_command(
    command: GGBTangentOnCircleCommand, lines: dict[str, GGBLine]
) -> None:
    lines[command.tangent_line].points.append(command.point_on_tangent)


def _add_vertex_to_bisector_from_angular_bisector_command(
    command: GGBAngularBisectorLinesCommand, lines: dict[str, GGBLine]
) -> None:
    lines[command.line_bisector].points.extend(command.corner)


def _add_tangency_elements_from_tangent_outside_circle_command(
    command: GGBTangentOutsideCircleCommand, lines: dict[str, GGBLine]
) -> None:
    lines[command.tangent_line1].points.append(command.point_on_tangent)
    lines[command.tangent_line2].points.append(command.point_on_tangent)
